"""`PhraseMatcher` ABC for other phrase-based spaczz matchers."""
import abc
import typing as ty
import warnings

from spacy.tokens import Doc
from spacy.vocab import Vocab

from .._search import PhraseSearcher
from ..customtypes import MatchResult
from ..customtypes import SpaczzType
from ..exceptions import KwargsWarning
from ..util import nest_defaultdict


class PhraseMatcher(abc.ABC):
    """Abstract base class for phrase-based matching in spaCy `Doc` objects."""

    name = "phrase_matcher"

    def __init__(self: "PhraseMatcher", vocab: Vocab, **defaults: ty.Any) -> None:
        """Initializes the matcher with the given defaults."""
        self.defaults = defaults
        self._type: SpaczzType = "phrase"
        self._callbacks: ty.Dict[str, PhraseCallback] = {}
        self._patterns: ty.DefaultDict[
            str, ty.DefaultDict[str, ty.Any]
        ] = nest_defaultdict(list)
        self._searcher = self._get_searcher(vocab)

    def __call__(self: "PhraseMatcher", doc: Doc) -> ty.List[MatchResult]:
        """Finds matches in `doc` given the matchers patterns."""
        matches = set()
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(  # noqa: B905
                patterns["patterns"], patterns["kwargs"]
            ):
                if not kwargs:
                    kwargs = self.defaults
                matches_wo_label = self._searcher.match(doc, pattern, **kwargs)
                if matches_wo_label:
                    matches_w_label = [
                        (label, *match_wo_label, str(pattern))
                        for match_wo_label in matches_wo_label
                    ]
                    for match in matches_w_label:
                        matches.add(match)

        sorted_matches = sorted(
            matches, key=lambda x: (-x[1], x[2] - x[1], x[3]), reverse=True
        )
        for i, (label, _start, _end, _ratio, _pattern) in enumerate(sorted_matches):
            on_match = self._callbacks.get(label)
            if on_match:
                on_match(self, doc, i, sorted_matches)
        return sorted_matches

    def __contains__(self: "PhraseMatcher", label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return label in self._patterns

    def __len__(self: "PhraseMatcher") -> int:
        """The number of labels added to the matcher."""
        return len(self._patterns)

    def __reduce__(
        self: "PhraseMatcher",
    ) -> ty.Tuple[ty.Any, ty.Any]:  # Precisely typing this would be really long.
        """Interface for pickling the matcher."""
        data = (
            self.__class__,
            self.vocab,
            self._patterns,
            self._callbacks,
            self.defaults,
        )
        return (unpickle_matcher, data)

    @property
    def labels(self: "PhraseMatcher") -> ty.Tuple[str, ...]:
        """All labels present in the matcher."""
        return tuple(self._patterns.keys())

    @property
    def patterns(self: "PhraseMatcher") -> ty.List[ty.Dict[str, ty.Any]]:
        """Get all patterns and kwargs that were added to the matcher."""
        all_patterns = []
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(  # noqa: B905
                patterns["patterns"], patterns["kwargs"]
            ):
                p = {"label": label, "pattern": pattern.text, "type": self.type}
                if kwargs:
                    p["kwargs"] = kwargs
                all_patterns.append(p)
        return all_patterns

    @property
    def type(self: "PhraseMatcher") -> SpaczzType:
        """Getter for the matchers `SpaczzType`."""
        return self._type

    @property
    def vocab(self: "PhraseMatcher") -> Vocab:
        """Getter for the matchers `Vocab`."""
        return self._searcher.vocab

    def add(
        self: "PhraseMatcher",
        label: str,
        patterns: ty.List[Doc],
        kwargs: ty.Optional[ty.List[ty.Dict[str, ty.Any]]] = None,
        on_match: "PhraseCallback" = None,
    ) -> None:
        """Add a rule to the matcher."""
        if not isinstance(patterns, list):
            raise TypeError("Patterns must be a list of `Doc` objects.")
        if kwargs is None:
            kwargs = [{} for _ in patterns]
        elif len(kwargs) < len(patterns):
            warnings.warn(
                """There are more patterns then there are kwargs.
                Patterns not matched to a kwarg dict will have default settings.""",
                KwargsWarning,
                stacklevel=2,
            )
            kwargs.extend([{} for _ in range(len(patterns) - len(kwargs))])
        elif len(kwargs) > len(patterns):
            warnings.warn(
                """There are more kwargs dicts than patterns.
                The extra kwargs will be ignored.""",
                KwargsWarning,
                stacklevel=2,
            )
        for pattern, kwarg in zip(patterns, kwargs):  # noqa: B905
            if isinstance(pattern, Doc):
                self._patterns[label]["patterns"].append(pattern)
            else:
                raise TypeError("`patterns` must be a list of `Doc` objects.")
            if isinstance(kwarg, dict):
                self._patterns[label]["kwargs"].append(kwarg)
            else:
                raise TypeError("`kwargs` must be a list of dicts.")
        self._callbacks[label] = on_match

    def remove(self: "PhraseMatcher", label: str) -> None:
        """Remove a label and its respective patterns from the matcher."""
        try:
            del self._patterns[label]
            del self._callbacks[label]
        except KeyError:
            raise ValueError(
                f"The label: '{label}' does not exist within "  # noqa: B907
                "the matcher rules."
            )

    @staticmethod
    @abc.abstractmethod
    def _get_searcher(vocab: Vocab) -> PhraseSearcher:
        """Initializes the searcher for this matcher."""
        pass  # pragma: no cover


PMT = ty.TypeVar("PMT", bound=PhraseMatcher)
PhraseCallback = ty.Optional[ty.Callable[[PMT, Doc, int, ty.List[MatchResult]], None]]


def unpickle_matcher(
    matcher: ty.Type[PhraseMatcher],
    vocab: Vocab,
    patterns: ty.DefaultDict[str, ty.DefaultDict[str, ty.Any]],
    callbacks: ty.Dict[str, PhraseCallback],
    defaults: ty.Any,
) -> PhraseMatcher:
    """Will return a matcher from pickle protocol."""
    matcher_instance = matcher(vocab, **defaults)
    for key, specs in patterns.items():
        callback = callbacks.get(key)
        matcher_instance.add(key, specs["patterns"], specs["kwargs"], on_match=callback)
    return matcher_instance



"""`FuzzySearcher` searches for phrase-based fuzzy matches in spaCy `Doc` objects."""
import typing as ty

from spacy.vocab import Vocab

from .phrasesearcher import PhraseSearcher
from ..customtypes import TextContainer
from ..registry.fuzzyfuncs import get_fuzzy_func


class FuzzySearcher(PhraseSearcher):
    """Class for phrase-based fuzzy match searching in spaCy `Doc` objects."""

    def __init__(self: "FuzzySearcher", vocab: Vocab) -> None:
        """Initializes the searcher."""
        super().__init__(vocab=vocab)

    def compare(
        self: "FuzzySearcher",
        s1: TextContainer,
        s2: TextContainer,
        *,
        ignore_case: bool = True,
        min_r: int = 0,
        fuzzy_func: str = "simple",
        **kwargs: ty.Any,
    ) -> int:
        """Peforms fuzzy matching between two spaCy container objects.

        spaCy containers are `Doc`, `Span`, or `Token` objects.

        Applies the given fuzzy matching algorithm (`fuzzy_func`)
        to two spaCy containers and returns the resulting fuzzy ratio.

        Args:
            s1: First spaCy container for comparison.
            s2: Second spaCy container for comparison.
            ignore_case: Whether to lower-case `s1` and `s2`
                before comparison or not. Default is `True`.
            min_r: Minimum ratio needed to match as a value between `0` and `100`.
                For ratio < `min_r`, `0` is returned instead.
                Default is `0`, which deactivates this behaviour.
            fuzzy_func: Key name of fuzzy matching function to use.
                Default is `"simple"`.
            **kwargs: Overflow for kwargs from parent class.

        Returns:
            The fuzzy ratio between `s1` and `s2`.

        Example:
            >>> import spacy
            >>> from spaczz._search import FuzzySearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = FuzzySearcher(nlp.vocab)
            >>> searcher.compare(nlp("spaczz"), nlp("spacy"))
            73
        """
        if ignore_case:
            s1_text = s1.text.lower()
            s2_text = s2.text.lower()
        else:
            s1_text = s1.text
            s2_text = s2.text
        return round(get_fuzzy_func(fuzzy_func)(s1_text, s2_text, score_cutoff=min_r))



"""Custom spaczz types."""
import typing as ty

try:
    from typing import Literal  # type: ignore
except ImportError:  # pragma: no cover
    from typing_extensions import Literal  # type: ignore

from spacy.tokens import Doc
from spacy.tokens import Span
from spacy.tokens import Token

DocLike = ty.Union[Doc, Span]
FlexLiteral = Literal["default", "min", "max"]
FlexType = ty.Union[int, FlexLiteral]  # type: ignore
TextContainer = ty.Union[Doc, Span, Token]
SearchResult = ty.Tuple[int, int, int]
MatchResult = ty.Tuple[str, int, int, int, str]
SpaczzType = Literal["fuzzy", "regex", "token", "similarity", "phrase"]
RulerPattern = ty.Dict[
    str, ty.Union[str, ty.Dict[str, ty.Any], ty.List[ty.Dict[str, ty.Any]]]
]
RulerResult = ty.Tuple[str, int, int, int, str, SpaczzType]  # type: ignore[valid-type]



"""Function and object registries."""
from .fuzzyfuncs import fuzzy_funcs
from .fuzzyfuncs import get_fuzzy_func
from .repatterns import get_re_pattern
from .repatterns import re_patterns
from .reweights import get_re_weights
from .reweights import re_weights

__all__ = [
    "fuzzy_funcs",
    "get_fuzzy_func",
    "get_re_pattern",
    "re_patterns",
    "get_re_weights",
    "re_weights",
]



"""Module for the SpaczzRuler."""
from collections import defaultdict
from pathlib import Path
import typing as ty
import warnings

from spacy.language import Language
from spacy.pipeline import Pipe
from spacy.scorer import get_ner_prf
from spacy.tokens import Doc
from spacy.tokens import Span
from spacy.training import Example
from spacy.training import validate_examples
from spacy.util import registry
from spacy.util import SimpleFrozenDict
from spacy.util import SimpleFrozenList
import srsly

from ..customtypes import RulerPattern
from ..customtypes import RulerResult
from ..customtypes import SpaczzType
from ..exceptions import PatternTypeWarning
from ..matcher import FuzzyMatcher
from ..matcher import RegexMatcher
from ..matcher import TokenMatcher
from ..util import ensure_path
from ..util import nest_defaultdict
from ..util import read_from_disk
from ..util import write_to_disk

DEFAULT_ENT_ID_SEP = "||"
SIMPLE_FROZEN_DICT = SimpleFrozenDict()
SIMPLE_FROZEN_LIST = SimpleFrozenList()

REQUIRE_PATTERNS_WARNING = (
    "The component 'spaczz_ruler' does not have any patterns defined."
)
warnings.filterwarnings("once", message=REQUIRE_PATTERNS_WARNING)

DEFAULT_CONFIG = {
    "overwrite_ents": False,
    "ent_id_sep": DEFAULT_ENT_ID_SEP,
    "fuzzy_defaults": SIMPLE_FROZEN_DICT,
    "regex_defaults": SIMPLE_FROZEN_DICT,
    "token_defaults": SIMPLE_FROZEN_DICT,
    "scorer": {"@scorers": "spaczz.spaczz_ruler_scorer.v1"},
}


def spaczz_ruler_scorer(
    examples: ty.Iterable[Example], **kwargs: ty.Any
) -> ty.Dict[str, ty.Any]:
    """Default spaczz scorer."""
    return get_ner_prf(examples)


if getattr(registry, "scorers", None):  # pragma: no cover

    @registry.scorers("spaczz.spaczz_ruler_scorer.v1")  # type: ignore[attr-defined]
    def make_spaczz_ruler_scorer() -> (
        ty.Callable[[ty.Iterable[Example]], ty.Dict[str, ty.Any]]
    ):
        """Wrapping `spaczz_ruler_scorer` in a callable."""
        return spaczz_ruler_scorer

else:
    del DEFAULT_CONFIG["scorer"]


@Language.factory(
    "spaczz_ruler",
    assigns=["doc.ents", "token.ent_type", "token.ent_iob"],
    default_config=DEFAULT_CONFIG,
    default_score_weights={
        "ents_f": 1.0,
        "ents_p": 0.0,
        "ents_r": 0.0,
        "ents_per_type": None,
    },
)
def make_spaczz_ruler(
    nlp: Language,
    name: str,
    overwrite_ents: bool,
    ent_id_sep: str,
    fuzzy_defaults: ty.Dict[str, ty.Any],
    regex_defaults: ty.Dict[str, ty.Any],
    token_defaults: ty.Dict[str, ty.Any],
    scorer: ty.Optional[ty.Callable] = None,
) -> "SpaczzRuler":
    """Factory method for creating a `SpaczzRuler`."""
    return SpaczzRuler(
        nlp,
        name,
        overwrite_ents=overwrite_ents,
        ent_id_sep=ent_id_sep,
        fuzzy_defaults=fuzzy_defaults,
        regex_defaults=regex_defaults,
        token_defaults=token_defaults,
        scorer=scorer,
    )


class SpaczzRuler(Pipe):
    """The `SpaczzRuler` adds fuzzy matches to spaCy `Doc.ents`.

    It can be combined with other spaCy NER components like the statistical
    `EntityRecognizer`, and/or the `EntityRuler` it is inspired by, to boost accuracy.
    After initialization, the component is typically added to the pipeline
    using `nlp.add_pipe`.

    Attributes:
        nlp (Language): The shared `Language` object that passes its `Vocab` to the
            matchers (not currently used by spaczz matchers) and processes fuzzy
            patterns.
        name (str): Instance name of the current pipeline component. Typically
            passed in automatically from the factory when the component is
            added. Used to disable the current entity ruler while creating
            phrase patterns with the nlp object.
        overwrite_ents (bool): If existing entities are present, e.g. entities
            added by the model, overwrite them by matches if necessary.
        ent_id_sep (str): Separator used internally for entity IDs.
        scorer (Optional[Callable]): The scoring method for the ruler.
        fuzzy_matcher (FuzzyMatcher): The `FuzzyMatcher` instance
            the spaczz ruler will use for fuzzy phrase matching.
        regex_matcher (RegexMatcher): The `RegexMatcher` instance
            the spaczz ruler will use for regex phrase matching.
        token_matcher (TokenMatcher): The `TokenMatcher` instance
            the spaczz ruler will use for fuzzy token matching.
        defaults (Dict[str, Any]): Default match settings for their respective matchers.
    """

    name = "spaczz_ruler"

    def __init__(
        self: "SpaczzRuler",
        nlp: Language,
        name: str = "spaczz_ruler",
        *,
        overwrite_ents: bool = False,
        ent_id_sep: str = DEFAULT_ENT_ID_SEP,
        fuzzy_defaults: ty.Dict[str, ty.Any] = SIMPLE_FROZEN_DICT,
        regex_defaults: ty.Dict[str, ty.Any] = SIMPLE_FROZEN_DICT,
        token_defaults: ty.Dict[str, ty.Any] = SIMPLE_FROZEN_DICT,
        patterns: ty.Optional[ty.List[RulerPattern]] = None,
        scorer: ty.Optional[ty.Callable] = spaczz_ruler_scorer,
    ) -> None:
        """Initialize the spaczz ruler.

        If `patterns` is supplied here, it needs to be an iterable of spaczz patterns:
        dictionaries with `"label"`, `"pattern"`, and `"type"` keys.
        If the patterns are fuzzy or regex phrase patterns
        they can include the optional `"kwargs"` keys.

        For example, a fuzzy phrase pattern::

            {
                'label': 'ORG',
                'pattern': 'Apple',
                'kwargs': {'min_r2': 90},
                'type': 'fuzzy',
            }

        Or, a token pattern::

            {
                'label': 'ORG',
                'pattern': [{'TEXT': {'FUZZY': 'Apple'}}],
                'type': 'token',
            }

        Args:
            nlp (Language): The shared `Language` object that passes its `Vocab` to the
                matchers (not currently used by spaczz matchers) and processes fuzzy
                patterns.
            name (str): Instance name of the current pipeline component. Typically
                passed in automatically from the factory when the component is
                added. Used to disable the current entity ruler while creating
                phrase patterns with the nlp object.
            overwrite_ents (bool): If existing entities are present, e.g. entities
                added by the model, overwrite them by matches if necessary.
            ent_id_sep (str): Separator used internally for entity IDs.
            fuzzy_defaults (Dict[str, Any]): Modified default parameters to use with
                the `FuzzyMatcher`. Default is `None`.
            regex_defaults (Dict[str, Any]): Modified default parameters to use with
                the `RegexMatcher`. Default is `None`.
            token_defaults (Dict[str, Any]): Modified default parameters to use with
                the `TokenMatcher`. Default is `None`.
            patterns (Optional[Dict[str, str | Dict[str, Any] | List[Dict[str, Any]]]):
                Optional patterns to load in.
            scorer (Optional[Callable]): The scoring method. Defaults to
                `spacy.scorer.get_ner_prf`.

        Raises:
            TypeError: If matcher defaults passed are not dictionaries.
        """
        self.nlp = nlp
        self.name = name
        self.overwrite = overwrite_ents
        self._fuzzy_patterns: ty.DefaultDict[
            str, ty.DefaultDict[str, ty.Any]
        ] = nest_defaultdict(list)
        self._regex_patterns: ty.DefaultDict[
            str, ty.DefaultDict[str, ty.Any]
        ] = nest_defaultdict(list)
        self._token_patterns: ty.DefaultDict[
            str, ty.List[ty.List[ty.Dict[str, ty.Any]]]
        ] = defaultdict(list)
        self.ent_id_sep = ent_id_sep
        self._ent_ids: ty.DefaultDict[ty.Any, ty.Any] = defaultdict(dict)
        self.defaults = {}
        default_names = ("fuzzy_defaults", "regex_defaults", "token_defaults")
        for default, name in zip(  # noqa: B905
            (fuzzy_defaults, regex_defaults, token_defaults), default_names
        ):
            if isinstance(default, dict):
                self.defaults[name] = default
            else:
                raise TypeError(
                    (
                        "`{name}` must be a dictionary of keyword arguments,",
                        f"not `{type(default)}`.",
                    )
                )
        self.fuzzy_matcher = FuzzyMatcher(nlp.vocab, **self.defaults["fuzzy_defaults"])
        self.regex_matcher = RegexMatcher(nlp.vocab, **self.defaults["regex_defaults"])
        self.token_matcher = TokenMatcher(nlp.vocab, **self.defaults["token_defaults"])
        if patterns is not None:
            self.add_patterns(patterns)
        self.scorer = scorer

    def __call__(self: "SpaczzRuler", doc: Doc) -> Doc:
        """Find matches in document and add them as entities.

        Args:
            doc: The `Doc` object in the pipeline.

        Returns:
            The `Doc` with added entities, if available.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> doc = nlp.make_doc("My name is Anderson, Grunt")
            >>> ruler.add_patterns([{"label": "NAME", "pattern": "Grant Andersen",
                "type": "fuzzy", "kwargs": {"fuzzy_func": "token_sort"}}])
            >>> doc = ruler(doc)
            >>> "Anderson, Grunt" in [ent.text for ent in doc.ents]
            True
        """
        error_handler = self.get_error_handler()
        try:
            matches = self.match(doc)
            self.set_annotations(doc, matches)
            return doc
        except Exception as e:
            error_handler(self.name, self, [doc], e)

    def __contains__(self: "SpaczzRuler", label: str) -> bool:
        """Whether a label is present in the patterns."""
        return (
            label in self._fuzzy_patterns
            or label in self._regex_patterns
            or label in self._token_patterns
        )

    def __len__(self: "SpaczzRuler") -> int:
        """The number of all patterns added to the ruler."""
        n_fuzzy_patterns = sum(
            len(p["patterns"]) for p in self._fuzzy_patterns.values()
        )
        n_regex_patterns = sum(
            len(p["patterns"]) for p in self._regex_patterns.values()
        )
        n_token_patterns = sum(len(p) for p in self._token_patterns.values())
        return n_fuzzy_patterns + n_regex_patterns + n_token_patterns

    @property
    def ent_ids(self: "SpaczzRuler") -> ty.Tuple[ty.Optional[str], ...]:
        """All entity ids present in the match patterns id properties.

        Returns:
            The unique string entity ids as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy", "id": "BEAT"}])
            >>> ruler.ent_ids
            ('BEAT',)
        """
        keys = set(self._fuzzy_patterns.keys())
        keys.update(self._regex_patterns.keys())
        keys.update(self._token_patterns.keys())
        all_ent_ids = set()

        for k in keys:
            if self.ent_id_sep in k:
                _, ent_id = self._split_label(k)
                all_ent_ids.add(ent_id)
        return tuple(all_ent_ids)

    @property
    def labels(self: "SpaczzRuler") -> ty.Tuple[str, ...]:
        """All labels present in the ruler.

        Returns:
            The unique string labels as a tuple.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> ruler.labels
            ('AUTHOR',)
        """
        keys = set(self._fuzzy_patterns.keys())
        keys.update(self._regex_patterns.keys())
        keys.update(self._token_patterns.keys())
        all_labels = set()
        for k in keys:
            if self.ent_id_sep in k:
                label, _ = self._split_label(k)
                all_labels.add(label)
            else:
                all_labels.add(k)
        return tuple(sorted(all_labels))

    @property
    def patterns(self: "SpaczzRuler") -> ty.List[RulerPattern]:
        """Get all patterns and kwargs that were added to the ruler.

        Returns:
            The original patterns and kwargs, one dictionary for each combination.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "STREET", "pattern": "street_addresses",
                "type": "regex", "kwargs": {"predef": True}}])
            >>> ruler.patterns == [
                {
                    "label": "STREET",
                    "pattern": "street_addresses",
                    "type": "regex",
                    "kwargs": {"predef": True},
                    },
                    ]
            True
        """
        all_patterns = []
        for label, fuzzy_patterns in self._fuzzy_patterns.items():
            for fuzzy_pattern, fuzzy_kwargs in zip(  # noqa: B905
                fuzzy_patterns["patterns"], fuzzy_patterns["kwargs"]
            ):
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": fuzzy_pattern.text, "type": "fuzzy"}
                if fuzzy_kwargs:
                    p["kwargs"] = fuzzy_kwargs
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        for label, regex_patterns in self._regex_patterns.items():
            for regex_pattern, regex_kwargs in zip(  # noqa: B905
                regex_patterns["patterns"], regex_patterns["kwargs"]
            ):
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": regex_pattern, "type": "regex"}
                if regex_kwargs:
                    p["kwargs"] = regex_kwargs
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        for label, token_patterns in self._token_patterns.items():
            for token_pattern in token_patterns:
                ent_label, ent_id = self._split_label(label)
                p = {"label": ent_label, "pattern": token_pattern, "type": "token"}
                if ent_id:
                    p["id"] = ent_id
                all_patterns.append(p)
        return all_patterns

    def add_patterns(
        self: "SpaczzRuler",
        patterns: ty.List[RulerPattern],
    ) -> None:
        """Add patterns to the ruler.

        A pattern must be a spaczz pattern:
        `{label (str), pattern (str or list), type (str),
        optional kwargs (dict[str, Any]), and optional id (str)}`.

        For example, a fuzzy phrase pattern::

            {
                'label': 'ORG',
                'pattern': 'Apple',
                'kwargs': {'min_r2': 90},
                'type': 'fuzzy',
            }

        Or, a token pattern::

            {
                'label': 'ORG',
                'pattern': [{'TEXT': {'FUZZY': 'Apple'}}],
                'type': 'token',
            }

        To utilize regex flags, use inline flags.

        Args:
            patterns: The spaczz patterns to add.

        Raises:
            TypeError: If `patterns` is not a list of dicts.
            ValueError: If one or more patterns do not conform
                the spaczz pattern structure.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> "AUTHOR" in ruler.labels
            True
        """
        # disable the nlp components after this one in case
        # they hadn't been initialized / deserialised yet
        try:
            current_index = -1
            for i, (_name, pipe) in enumerate(self.nlp.pipeline):
                if self == pipe:
                    current_index = i
                    break
            subsequent_pipes = [
                pipe for pipe in self.nlp.pipe_names[current_index + 1 :]
            ]
        except ValueError:
            subsequent_pipes = []
        with self.nlp.select_pipes(disable=subsequent_pipes):
            token_patterns = []
            fuzzy_pattern_labels = []
            fuzzy_pattern_texts = []
            fuzzy_pattern_kwargs = []
            fuzzy_pattern_ids = []
            regex_pattern_labels = []
            regex_pattern_texts = []
            regex_pattern_kwargs = []
            regex_pattern_ids = []

            for entry in patterns:
                try:
                    if isinstance(entry, dict):
                        if entry["type"] == "fuzzy":
                            fuzzy_pattern_labels.append(entry["label"])
                            fuzzy_pattern_texts.append(entry["pattern"])
                            fuzzy_pattern_kwargs.append(entry.get("kwargs", {}))
                            fuzzy_pattern_ids.append(entry.get("id"))
                        elif entry["type"] == "regex":
                            regex_pattern_labels.append(entry["label"])
                            regex_pattern_texts.append(entry["pattern"])
                            regex_pattern_kwargs.append(entry.get("kwargs", {}))
                            regex_pattern_ids.append(entry.get("id"))
                        elif entry["type"] == "token":
                            token_patterns.append(entry)
                        else:
                            warnings.warn(
                                f"""Spaczz pattern "type" must be "fuzzy", "regex",
                                or "token", not {entry["type"]}. Skipping this pattern.
                                """,
                                PatternTypeWarning,
                                stacklevel=2,
                            )
                    else:
                        raise TypeError(("Patterns must be a list of dicts."))
                except KeyError:
                    raise ValueError(
                        (
                            "One or more patterns do not conform",
                            "to spaczz pattern structure: ",
                            "{label (str), pattern (str or list), type (str),",
                            "optional kwargs (dict[str, Any]),",
                            "and optional id (str)}.",
                        )
                    )

            fuzzy_patterns = []
            for flabel, fpattern, fkwargs, fent_id in zip(  # noqa: B905
                fuzzy_pattern_labels,
                self.nlp.pipe(ty.cast(str, fuzzy_pattern_texts)),
                fuzzy_pattern_kwargs,
                fuzzy_pattern_ids,
            ):
                fuzzy_pattern = {
                    "label": flabel,
                    "pattern": fpattern,
                    "kwargs": fkwargs,
                    "type": "fuzzy",
                }
                if fent_id:
                    fuzzy_pattern["id"] = fent_id
                fuzzy_patterns.append(fuzzy_pattern)

            regex_patterns = []
            for rlabel, rpattern, rkwargs, rent_id in zip(  # noqa: B905
                regex_pattern_labels,
                regex_pattern_texts,
                regex_pattern_kwargs,
                regex_pattern_ids,
            ):
                regex_pattern = {
                    "label": rlabel,
                    "pattern": rpattern,
                    "kwargs": rkwargs,
                    "type": "regex",
                }
                if rent_id:
                    regex_pattern["id"] = rent_id
                regex_patterns.append(regex_pattern)

            self._add_patterns(fuzzy_patterns, regex_patterns, token_patterns)

    def clear(self: "SpaczzRuler") -> None:
        """Reset all patterns."""
        self._fuzzy_patterns = nest_defaultdict(list)
        self._regex_patterns = nest_defaultdict(list)
        self._token_patterns = defaultdict(list)
        self._ent_ids = defaultdict(dict)
        self.fuzzy_matcher = FuzzyMatcher(
            self.nlp.vocab, **self.defaults["fuzzy_defaults"]
        )
        self.regex_matcher = RegexMatcher(
            self.nlp.vocab, **self.defaults["regex_defaults"]
        )
        self.token_matcher = TokenMatcher(
            self.nlp.vocab, **self.defaults["token_defaults"]
        )

    def initialize(
        self: "SpaczzRuler",
        get_examples: ty.Callable[[], ty.Iterable[Example]],
        *,
        nlp: ty.Optional[Language] = None,
        patterns: ty.Optional[ty.Sequence[RulerPattern]] = None,
    ) -> None:
        """Initialize the pipe for training.

        Args:
            get_examples: Function that returns a representative sample
                of gold-standard Example objects.
            nlp: The current nlp object the component is part of.
            patterns: The list of patterns.
        """
        self.clear()
        if patterns:
            self.add_patterns(patterns)  # type: ignore[arg-type]

    def match(self: "SpaczzRuler", doc: Doc) -> ty.List[RulerResult]:
        """Used in call to find matches in `doc`."""
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="\\[W036")
            matches: ty.List[ty.Tuple[str, int, int, int, str, SpaczzType]] = (
                [(*match, "fuzzy") for match in self.fuzzy_matcher(doc)]
                + [(*match, "regex") for match in self.regex_matcher(doc)]
                + [(*match, "token") for match in self.token_matcher(doc)]
            )
        matches = self._get_final_matches(matches)
        return sorted(matches, key=lambda x: (x[2] - x[1], -x[1], x[3]), reverse=True)

    def remove(self: "SpaczzRuler", ent_id: str) -> None:
        """Remove patterns by their `ent_id`."""
        label_id_pairs = [
            (label, eid) for (label, eid) in self._ent_ids.values() if eid == ent_id
        ]
        if not label_id_pairs:
            raise ValueError(
                f"The `ent_id`: '{ent_id}' does not exist within the ruler."  # noqa: B907
            )
        created_labels = [
            self._create_label(label, eid) for (label, eid) in label_id_pairs
        ]
        # remove the patterns from self.fuzzy_patterns
        self._fuzzy_patterns = nest_defaultdict(
            list,
            1,
            {
                label: val
                for (label, val) in self._fuzzy_patterns.items()
                if label not in created_labels
            },
        )
        # remove the patterns from self.regex_patterns
        self._regex_patterns = nest_defaultdict(
            list,
            1,
            {
                label: val
                for (label, val) in self._regex_patterns.items()
                if label not in created_labels
            },
        )
        # remove the patterns from self.token_pattern
        self._token_patterns = defaultdict(
            list,
            {
                label: val
                for (label, val) in self._token_patterns.items()
                if label not in created_labels
            },
        )
        # remove the patterns from the matchers
        for label in created_labels:
            if label in self.fuzzy_matcher:
                self.fuzzy_matcher.remove(label)
            elif label in self.regex_matcher:
                self.regex_matcher.remove(label)
            else:
                self.token_matcher.remove(label)

    def score(
        self: "SpaczzRuler", examples: ty.Iterable[Example], **kwargs: ty.Any
    ) -> ty.Any:
        """Pipeline scoring for spaCy >= 3.0, < 3.2 compatibility."""
        validate_examples(examples, "SpaczzRuler.score")
        return get_ner_prf(examples)

    def set_annotations(
        self: "SpaczzRuler",
        doc: Doc,
        matches: ty.List[ty.Tuple[str, int, int, int, str, SpaczzType]],
    ) -> None:
        """Modify the document in place."""
        entities = list(doc.ents)
        new_entities = []
        seen_tokens: ty.Set[int] = set()
        for match_id, start, end, ratio, pattern, match_type in matches:
            if any(t.ent_type for t in doc[start:end]) and not self.overwrite:
                continue
            # check for end - 1 here because boundaries are inclusive
            if start not in seen_tokens and end - 1 not in seen_tokens:
                if match_id in self._ent_ids:
                    label, ent_id = self._ent_ids[match_id]
                    span = Span(doc, start, end, label=label)
                    if ent_id:
                        for token in span:
                            token.ent_id_ = ent_id
                else:
                    span = Span(doc, start, end, label=match_id)
                span = self._update_custom_attrs(
                    span,
                    match_id=match_id,
                    ratio=ratio,
                    pattern=pattern,
                    match_type=match_type,
                )
                new_entities.append(span)
                entities = [
                    e for e in entities if not (e.start < end and e.end > start)
                ]
                seen_tokens.update(range(start, end))
        doc.ents = tuple(entities + new_entities)  # type: ignore

    def from_bytes(
        self: "SpaczzRuler",
        patterns_bytes: bytes,
        *,
        exclude: ty.Iterable[str] = SIMPLE_FROZEN_LIST,
    ) -> "SpaczzRuler":
        """Load the spaczz ruler from a bytestring.

        Args:
            patterns_bytes: The bytestring to load.
            exclude: For spaCy consistency.

        Returns:
            The loaded spaczz ruler.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> ruler_bytes = ruler.to_bytes()
            >>> new_ruler = SpaczzRuler(nlp)
            >>> new_ruler = new_ruler.from_bytes(ruler_bytes)
            >>> "AUTHOR" in new_ruler
            True
        """
        cfg = srsly.msgpack_loads(patterns_bytes)
        self.clear()
        if isinstance(cfg, dict):
            self.add_patterns(cfg.get("patterns", cfg))
            self.defaults = cfg.get("defaults", {})
            if self.defaults.get("fuzzy_defaults"):
                self.fuzzy_matcher = FuzzyMatcher(
                    self.nlp.vocab, **self.defaults["fuzzy_defaults"]
                )
            if self.defaults.get("regex_defaults"):
                self.regex_matcher = RegexMatcher(
                    self.nlp.vocab, **self.defaults["regex_defaults"]
                )
            if self.defaults.get("token_defaults"):
                self.token_matcher = TokenMatcher(
                    self.nlp.vocab, **self.defaults["token_defaults"]
                )
            self.overwrite = cfg.get("overwrite", False)
            self.ent_id_sep = cfg.get("ent_id_sep", DEFAULT_ENT_ID_SEP)
        else:
            self.add_patterns(cfg)
        return self

    def to_bytes(
        self: "SpaczzRuler", *, exclude: ty.Iterable[str] = SIMPLE_FROZEN_LIST
    ) -> bytes:
        """Serialize the spaczz ruler patterns to a bytestring.

        Args:
            exclude: For spaCy consistency.

        Returns:
            The serialized patterns.

        Example:
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> ruler_bytes = ruler.to_bytes()
            >>> isinstance(ruler_bytes, bytes)
            True
        """
        serial = {
            "overwrite": self.overwrite,
            "ent_id_sep": self.ent_id_sep,
            "patterns": self.patterns,
            "defaults": self.defaults,
        }
        return srsly.msgpack_dumps(serial)

    def from_disk(
        self: "SpaczzRuler",
        path: ty.Union[str, Path],
        *,
        exclude: ty.Iterable[str] = SIMPLE_FROZEN_LIST,
    ) -> "SpaczzRuler":
        """Load the spaczz ruler from a file.

        Expects a file containing newline-delimited JSON (JSONL)
        with one entry per line.

        Args:
            path: The JSONL file to load.
            exclude: For spaCy consistency.

        Returns:
            The loaded spaczz ruler.

        Raises:
            ValueError: If `path` does not exist or cannot be accessed.

        Example:
            >>> import os
            >>> import tempfile
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            >>>     ruler.to_disk(f"{tmpdir}/ruler")
            >>>     new_ruler = SpaczzRuler(nlp)
            >>>     new_ruler = new_ruler.from_disk(f"{tmpdir}/ruler")
            >>> "AUTHOR" in new_ruler
            True
        """
        path = ensure_path(path)
        self.clear()
        depr_patterns_path = path.with_suffix(".jsonl")
        if path.suffix == ".jsonl":  # user provides a jsonl
            if path.is_file():
                patterns = srsly.read_jsonl(path)
                self.add_patterns(patterns)
            else:
                raise ValueError(
                    f"Couldn't read SpaczzRuler from '{path}'. "  # noqa: B907
                    "This file doesn't exist."
                )
        elif depr_patterns_path.is_file():
            patterns = srsly.read_jsonl(depr_patterns_path)
            self.add_patterns(patterns)
        elif path.is_dir():
            cfg = {}
            deserializers_patterns = {
                "patterns": lambda p: self.add_patterns(
                    srsly.read_jsonl(p.with_suffix(".jsonl"))
                )
            }
            deserializers_cfg = {"cfg": lambda p: cfg.update(srsly.read_json(p))}
            read_from_disk(path, deserializers_cfg, {})
            self.overwrite = cfg.get("overwrite", False)
            self.defaults = cfg.get("defaults", {})
            if self.defaults.get("fuzzy_defaults"):
                self.fuzzy_matcher = FuzzyMatcher(
                    self.nlp.vocab, **self.defaults["fuzzy_defaults"]
                )
            if self.defaults.get("regex_defaults"):
                self.regex_matcher = RegexMatcher(
                    self.nlp.vocab, **self.defaults["regex_defaults"]
                )
            if self.defaults.get("token_defaults"):
                self.token_matcher = TokenMatcher(
                    self.nlp.vocab, **self.defaults["token_defaults"]
                )
            self.ent_id_sep = cfg.get("ent_id_sep", DEFAULT_ENT_ID_SEP)
            read_from_disk(path, deserializers_patterns, {})
        else:  # path is not a valid directory or file
            raise ValueError(f"Could not access f{path}.")
        return self

    def to_disk(
        self: "SpaczzRuler",
        path: ty.Union[str, Path],
        *,
        exclude: ty.Iterable[str] = SIMPLE_FROZEN_LIST,
    ) -> None:
        """Save the spaczz ruler patterns to a directory.

        The patterns will be saved as newline-delimited JSON (JSONL).

        Args:
            path: The JSONL file to save.
            exclude: For spaCy consistency.

        Example:
            >>> import os
            >>> import tempfile
            >>> import spacy
            >>> from spaczz.pipeline import SpaczzRuler
            >>> nlp = spacy.blank("en")
            >>> ruler = SpaczzRuler(nlp)
            >>> ruler.add_patterns([{"label": "AUTHOR", "pattern": "Kerouac",
                "type": "fuzzy"}])
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            >>>     ruler.to_disk(f"{tmpdir}/ruler")
            >>>     isdir = os.path.isdir(f"{tmpdir}/ruler")
            >>> isdir
            True
        """
        path = ensure_path(path)
        cfg = {
            "overwrite": self.overwrite,
            "defaults": self.defaults,
            "ent_id_sep": self.ent_id_sep,
        }
        serializers = {
            "patterns": lambda p: srsly.write_jsonl(
                p.with_suffix(".jsonl"), self.patterns
            ),
            "cfg": lambda p: srsly.write_json(p, cfg),
        }
        if path.suffix == ".jsonl":  # user wants to save only JSONL
            srsly.write_jsonl(path, self.patterns)
        else:
            write_to_disk(path, serializers, {})

    def _add_patterns(
        self: "SpaczzRuler",
        fuzzy_patterns: ty.List[ty.Dict[str, ty.Any]],
        regex_patterns: ty.List[ty.Dict[str, ty.Any]],
        token_patterns: ty.List[ty.Dict[str, ty.Any]],
    ) -> None:
        """Helper function for add_patterns."""
        for entry in fuzzy_patterns + regex_patterns + token_patterns:
            label = entry["label"]
            if "id" in entry:
                ent_label = label
                label = self._create_label(label, entry["id"])
                self._ent_ids[label] = (ent_label, entry["id"])
            pattern = entry["pattern"]
            if isinstance(pattern, Doc):
                self._fuzzy_patterns[label]["patterns"].append(pattern)
                self._fuzzy_patterns[label]["kwargs"].append(entry["kwargs"])
            elif isinstance(pattern, str):
                self._regex_patterns[label]["patterns"].append(pattern)
                self._regex_patterns[label]["kwargs"].append(entry["kwargs"])
            elif isinstance(pattern, list):
                self._token_patterns[label].append(pattern)
            else:
                raise ValueError(
                    (
                        "One or more patterns do not conform",
                        "to spaczz pattern structure:",
                        "{label (str), pattern (str or list), type (str),",
                        "optional kwargs (dict[str, Any]),",
                        "and optional id (str)}.",
                    )
                )
        for label, patterns in self._fuzzy_patterns.items():
            self.fuzzy_matcher.add(label, patterns["patterns"], patterns["kwargs"])
        for label, patterns in self._regex_patterns.items():
            self.regex_matcher.add(label, patterns["patterns"], patterns["kwargs"])
        for label, token_patterns_ in self._token_patterns.items():
            self.token_matcher.add(label, token_patterns_)

    def _create_label(
        self: "SpaczzRuler", label: str, ent_id: ty.Union[str, None]
    ) -> str:
        """Join Entity label with ent_id if the pattern has an id attribute.

        Args:
            label: The entity label.
            ent_id: The optional entity id.

        Returns:
            The label and ent_id joined with configured ent_id_sep.
        """
        if isinstance(ent_id, str):
            label = f"{label}{self.ent_id_sep}{ent_id}"
        return label

    def _require_patterns(self: "SpaczzRuler") -> None:
        """Raise a warning if this component has no patterns defined."""
        if len(self) == 0:
            warnings.warn(
                REQUIRE_PATTERNS_WARNING,
                stacklevel=2,
            )

    def _split_label(
        self: "SpaczzRuler", label: str
    ) -> ty.Tuple[str, ty.Optional[str]]:
        """Split Entity label into ent_label and ent_id if it contains self.ent_id_sep.

        Args:
            label: The value of label in a pattern entry

        Returns:
            The separated ent_label and optional ent_id.
        """
        if self.ent_id_sep in label:
            ent_label, ent_id = label.rsplit(self.ent_id_sep, 1)
            return ent_label, ent_id
        else:
            ent_label = label
            return ent_label, None

    @staticmethod
    def _get_final_matches(
        matches: ty.List[RulerResult],
    ) -> ty.List[RulerResult]:
        final_matches = []
        lookup: ty.Dict[ty.Tuple[str, int, int], int] = dict()
        for match in matches:
            ratio = lookup.get(match[:3], 0)
            if match[3] > ratio:
                final_matches.append(match)
                lookup[match[:3]] = ratio
        return final_matches

    @staticmethod
    def _update_custom_attrs(
        span: Span, match_id: str, ratio: int, pattern: str, match_type: SpaczzType
    ) -> Span:
        """Update custom attributes for matches."""
        for token in span:
            token._.spaczz_token = True
            token._.spaczz_ratio = ratio
            token._.spaczz_pattern = pattern
            token._.spaczz_type = match_type
        return span



"""Registry of fuzzy matching functions."""
from functools import lru_cache

import catalogue
from rapidfuzz import fuzz

fuzzy_funcs = catalogue.create("spaczz", "fuzzy_funcs", entry_points=True)
fuzzy_funcs.register("simple", func=fuzz.ratio)
fuzzy_funcs.register("partial", func=fuzz.partial_ratio)
fuzzy_funcs.register("token", func=fuzz.token_ratio)
fuzzy_funcs.register("token_set", func=fuzz.token_set_ratio)
fuzzy_funcs.register("token_sort", func=fuzz.token_sort_ratio)
fuzzy_funcs.register("partial_token", func=fuzz.partial_token_ratio)
fuzzy_funcs.register("partial_token_set", func=fuzz.partial_token_set_ratio)
fuzzy_funcs.register("partial_token_sort", func=fuzz.partial_token_sort_ratio)
fuzzy_funcs.register("weighted", func=fuzz.WRatio)
fuzzy_funcs.register("quick", func=fuzz.QRatio)

try:
    from rapidfuzz.fuzz import partial_ratio_alignment

    fuzzy_funcs.register("partial_alignment", func=partial_ratio_alignment)
except ImportError:  # pragma: no cover
    pass

get_fuzzy_func = lru_cache(None)(fuzzy_funcs.get)



"""`FuzzyMatcher` with an API semi-analogous to spaCy's `PhraseMatcher`."""
import typing as ty

from spacy.tokens import Doc
from spacy.vocab import Vocab

from ._phrasematcher import PhraseCallback
from ._phrasematcher import PhraseMatcher
from .._search import FuzzySearcher
from ..customtypes import MatchResult
from ..customtypes import SpaczzType


class FuzzyMatcher(PhraseMatcher):
    """spaCy-like matcher for finding fuzzy phrase matches in `Doc` objects.

    Fuzzy matches patterns against the `Doc` it is called on.
    Accepts labeled patterns in the form of `Doc` objects with optional,
    per-pattern match settings.

    Attributes:
        name (str): Class attribute - the name of the matcher.
        defaults (dict[str, bool|int|str|Literal['default', 'min', 'max']]):
            Keyword arguments to be used as default match settings.
            Per-pattern match settings take precedence over defaults.

    Match Settings:
        ignore_case (bool): Whether to lower-case text before matching.
            Default is `True`.
        min_r (int): Minimum match ratio required.
        thresh (int): If this ratio is exceeded in initial scan,
            and `flex > 0`, no optimization will be attempted.
            If `flex == 0`, `thresh` has no effect. Default is `100`.
        fuzzy_func (str): Key name of fuzzy matching function to use.
            All rapidfuzz matching functions with default settings are available.
            Additional fuzzy matching functions can be registered by users.
            Included functions are:

            * `"simple"` = `ratio`
            * `"partial"` = `partial_ratio`
            * `"token"` = `token_ratio`
            * `"token_set"` = `token_set_ratio`
            * `"token_sort"` = `token_sort_ratio`
            * `"partial_token"` = `partial_token_ratio`
            * `"partial_token_set"` = `partial_token_set_ratio`
            * `"partial_token_sort"` = `partial_token_sort_ratio`
            * `"weighted"` = `WRatio`
            * `"quick"` = `QRatio`
            * `"partial_alignment"` = `partial_ratio_alignment`
              (Requires `rapidfuzz>=2.0.3`)

            Default is `"simple"`.
        flex (int|Literal['default', 'min', 'max']): Number of tokens to move match
            boundaries left and right during optimization. Can be an `int` with a max of
            `len(pattern)` and a min of `0`, (will warn and change if higher or lower).
            `"max"`, `"min"`, or `"default"` are also valid.
            Default is `"default"`: `len(pattern) // 2`.
        min_r1 (int|None): Optional granular control over the minimum match ratio
            required for selection during the initial scan.
            If `flex == 0`, `min_r1` will be overwritten by `min_r2`.
            If `flex > 0`, `min_r1` must be lower than `min_r2` and "low" in general
            because match boundaries are not flexed initially.
            Default is `None`, which will result in `min_r1` being set to
            `round(min_r / 1.5)`.
        min_r2 (int|None): Optional granular control over the minimum match ratio
            required for selection during match optimization.
            Needs to be higher than `min_r1` and "high" in general to ensure only
            quality matches are returned. Default is `None`, which will result in
            `min_r2` being set to `min_r`.
    """

    name = "fuzzy_matcher"

    def __init__(self: "FuzzyMatcher", vocab: Vocab, **defaults: ty.Any) -> None:
        """Initializes the matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object. Purely for consistency between spaCy
                and spaczz matcher APIs for now. spaczz matchers are currently pure
                Python and do not share vocabulary with spacy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings for the class instance.
        """
        super().__init__(vocab=vocab, **defaults)
        self._type: SpaczzType = "fuzzy"
        self._searcher = self._get_searcher(vocab)

    def __call__(self: "FuzzyMatcher", doc: Doc) -> ty.List[MatchResult]:
        """Finds matches in `doc` given the matchers patterns.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of `MatchResult` tuples,
            (label, start index, end index, match ratio, pattern).

        Example:
            >>> import spacy
            >>> from spaczz.matcher import FuzzyMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = FuzzyMatcher(nlp.vocab)
            >>> doc = nlp("Rdley Scott was the director of Alien.")
            >>> matcher.add("NAME", [nlp.make_doc("Ridley Scott")])
            >>> matcher(doc)
            [('NAME', 0, 2, 96, 'Ridley Scott')]
        """
        return super().__call__(doc)

    def __contains__(self: "FuzzyMatcher", label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return super().__contains__(label)

    def __len__(self: "FuzzyMatcher") -> int:
        """The number of labels added to the matcher."""
        return super().__len__()

    def __reduce__(
        self: "FuzzyMatcher",
    ) -> ty.Tuple[ty.Any, ty.Any]:  # Precisely typing this would be really long.
        """Interface for pickling the matcher."""
        return super().__reduce__()

    @property
    def labels(self: "FuzzyMatcher") -> ty.Tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique labels as a tuple of strings.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import FuzzyMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = FuzzyMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [nlp.make_doc("Kerouac")])
            >>> matcher.labels
            ('AUTHOR',)
        """
        return super().labels

    @property
    def patterns(self: "FuzzyMatcher") -> ty.List[ty.Dict[str, ty.Any]]:
        """Get all patterns and match settings that were added to the matcher.

        Returns:
            The patterns and their respective match settings as a list of dicts.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import FuzzyMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = FuzzyMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [nlp.make_doc("Kerouac")],
                [{"ignore_case": False}])
            >>> matcher.patterns == [
                {
                    "label": "AUTHOR",
                    "pattern": "Kerouac",
                    "type": "fuzzy",
                    "kwargs": {"ignore_case": False}
                    },
                    ]
            True
        """
        return super().patterns

    @property
    def type(self: "FuzzyMatcher") -> SpaczzType:
        """Getter for the matchers `SpaczzType`."""
        return super().type

    @property
    def vocab(self: "FuzzyMatcher") -> Vocab:
        """Getter for the matchers `Vocab`."""
        return super().vocab

    def add(
        self: "FuzzyMatcher",
        label: str,
        patterns: ty.List[Doc],
        kwargs: ty.Optional[ty.List[ty.Dict[str, ty.Any]]] = None,
        on_match: PhraseCallback = None,
    ) -> None:
        """Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of `Doc` objects and if `kwargs` is not `None`,
        `kwargs` must be a list of dicts.

        Args:
            label: Name of the rule added to the matcher.
            patterns: `Doc` objects that will be matched
                against the `Doc` object the matcher is called on.
            kwargs: Optional settings to modify the matching behavior.
                If supplying `kwargs`, one per pattern should be included.
                Empty dicts will use the matcher instances default settings.
                Default is `None`.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Warnings:
            KwargsWarning:
                * If there are more patterns than kwargs
                  default matching settings will be used
                  for extra patterns.
                * If there are more kwargs dicts than patterns,
                  the extra kwargs will be ignored.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import FuzzyMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = FuzzyMatcher(nlp.vocab)
            >>> matcher.add("SOUND", [nlp.make_doc("mooo")])
            >>> "SOUND" in matcher
            True
        """
        super().add(label, patterns=patterns, kwargs=kwargs, on_match=on_match)

    def remove(self: "FuzzyMatcher", label: str) -> None:
        """Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import FuzzyMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = FuzzyMatcher(nlp.vocab)
            >>> matcher.add("SOUND", [nlp.make_doc("mooo")])
            >>> matcher.remove("SOUND")
            >>> "SOUND" in matcher
            False
        """
        super().remove(label)

    @staticmethod
    def _get_searcher(vocab: Vocab) -> FuzzySearcher:
        """Initializes the searcher for this matcher."""
        return FuzzySearcher(vocab)



"""`SimilaritySearcher` searches for phrase-based sim matches in spaCy `Doc` objects."""
import typing as ty
import warnings

from spacy.vocab import Vocab

from .phrasesearcher import PhraseSearcher
from ..customtypes import TextContainer
from ..exceptions import MissingVectorsWarning


class SimilaritySearcher(PhraseSearcher):
    """Class for phrase-based similarity match searching in spaCy `Doc` objects."""

    def __init__(self: "SimilaritySearcher", vocab: Vocab) -> None:
        """Initializes a the searcher."""
        super().__init__(vocab=vocab)
        if vocab.vectors.n_keys == 0:
            warnings.warn(
                """The spaCy `Vocab` object has no word vectors.
                Similarity results may not be useful.""",
                MissingVectorsWarning,
                stacklevel=2,
            )

    def compare(
        self: "SimilaritySearcher",
        s1: TextContainer,
        s2: TextContainer,
        **kwargs: ty.Any,
    ) -> int:
        """Peforms similarity matching between two spaCy container objects.

        spaCy containers are `Doc`, `Span`, or `Token` objects.

        Args:
            s1: First spaCy container for comparison.
            s2: Second spaCy container for comparison.
            **kwargs: Overflow for kwargs from parent class.

        Returns:
            The similarity score between `s1` and `s2`.

        Example:
            >>> import spacy
            >>> from spaczz._search import SimilaritySearcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> searcher = SimilaritySearcher(nlp.vocab)
            >>> doc1 = nlp("I like apples.")
            >>> doc2 = nlp("I like grapes.")
            >>> searcher.compare(doc1, doc2) > 0
            True
        """
        return round(s1.similarity(s2) * 100)



"""`PhraseSearcher` ABC for other phrase-based spaczz searchers."""
import abc
import typing as ty
import warnings

from spacy.tokens import Doc
from spacy.vocab import Vocab

from .searchutil import filter_overlapping_matches
from ..customtypes import DocLike
from ..customtypes import FlexType
from ..customtypes import SearchResult
from ..customtypes import TextContainer
from ..exceptions import FlexWarning
from ..exceptions import RatioWarning


class PhraseSearcher(abc.ABC):
    """Abstract base class for phrase-based match searching in spaCy `Doc` objects."""

    def __init__(self: "PhraseSearcher", vocab: Vocab) -> None:
        """Initializes the searcher."""
        self.vocab = vocab

    @abc.abstractmethod
    def compare(
        self: "PhraseSearcher", s1: TextContainer, s2: TextContainer, **kwargs: ty.Any
    ) -> int:
        """Abstract method."""
        pass  # pragma: no cover

    def match(
        self: "PhraseSearcher",
        doc: Doc,
        query: DocLike,
        *,
        min_r: int = 75,
        thresh: int = 100,
        min_r1: ty.Optional[int] = None,
        min_r2: ty.Optional[int] = None,
        flex: FlexType = "default",
        **kwargs: ty.Any,
    ) -> ty.List[SearchResult]:
        """Returns phrase matches in a `Doc` object."""
        flex = self._calc_flex(query, flex)
        min_r1_, min_r2_ = self._set_ratios(min_r, min_r1, min_r2)
        min_r1_, min_r2_, thresh = self._check_ratios(min_r1_, min_r2_, thresh, flex)
        match_map = self._scan(doc, query, min_r1=min_r1_, **kwargs)

        if match_map:
            positions = list(match_map.keys())
            matches_w_nones = [
                self._optimize(
                    doc,
                    query,
                    match_values=match_map,
                    pos=pos,
                    flex=flex,
                    min_r2=min_r2_,
                    thresh=thresh,
                    **kwargs,
                )
                for pos in positions
            ]

            matches = [match for match in matches_w_nones if match]
            if matches:
                return filter_overlapping_matches(
                    sorted(
                        [match for match in matches_w_nones if match],
                        key=lambda x: (-x[2], x[0]),
                    ),
                )
            else:
                return []

        return []

    def _optimize(
        self: "PhraseSearcher",
        doc: Doc,
        query: DocLike,
        match_values: ty.Mapping[int, int],
        pos: int,
        flex: int,
        min_r2: int,
        thresh: int,
        **kwargs: ty.Any,
    ) -> ty.Optional[ty.Tuple[int, int, int]]:
        """Optimizes a potential match by flexing match boundaries.

        For a match from `._scan` that has match ratio >= `min_r1`, the match boundaries
        will be extended both left and right by `flex` number of tokens and matched
        back to the original `query`.
        The optimal start index, end index, and match ratio are returned
        if the match ratio >= `min_r2`.

        Args:
            doc: `Doc` to optimize initial matches over.
            query: `Doc` or `Span` to match against `doc`.
            match_values: Dict of initial match start indices keys to match ratios
                values.
            pos: Start index of the match being optimized.
            flex: Number of tokens to move match boundaries
                left and right during match optimization.
            min_r2: Minimum match ratio required
                to pass optimization. This should be high enough
                to only return quality matches.
            thresh: If this ratio <= match ratio from the initial scan,
                no optimization will be attempted.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            Start index, end index, match ratio tuple, or `None`.
        """
        doc_len = len(doc)
        p_l, bp_l = [pos] * 2
        p_r, bp_r = [pos + len(query)] * 2
        r = match_values[pos]
        if flex and not r >= thresh:
            optim_r = r
            for f in range(1, flex + 1):
                if p_l - f >= 0:
                    new_r = self.compare(
                        query, doc[p_l - f : p_r], min_r=optim_r, **kwargs
                    )
                    if new_r:
                        optim_r = new_r
                        bp_l = p_l - f
                        bp_r = p_r
                if p_l + f < p_r:
                    new_r = self.compare(
                        query, doc[p_l + f : p_r], min_r=optim_r, **kwargs
                    )
                    if new_r:
                        optim_r = new_r
                        bp_l = p_l + f
                        bp_r = p_r
                if p_r - f > p_l:
                    new_r = self.compare(
                        query, doc[p_l : p_r - f], min_r=optim_r, **kwargs
                    )
                    if new_r:
                        optim_r = new_r
                        bp_l = p_l
                        bp_r = p_r - f
                if p_r + f <= doc_len:
                    new_r = self.compare(
                        query, doc[p_l : p_r + f], min_r=optim_r, **kwargs
                    )
                    if new_r:
                        optim_r = new_r
                        bp_l = p_l
                        bp_r = p_r + f
                if p_l - f >= 0 and p_r + f <= doc_len:
                    new_r = self.compare(
                        query, doc[p_l - f : p_r + f], min_r=optim_r, **kwargs
                    )
                    if new_r:
                        optim_r = new_r
                        bp_l = p_l - f
                        bp_r = p_r + f
                if p_l + f < p_r and p_r - f > p_l:
                    new_r = self.compare(
                        query, doc[p_l + f : p_r - f], min_r=optim_r, **kwargs
                    )
                    if new_r:
                        optim_r = new_r
                        bp_l = p_l + f
                        bp_r = p_r - f
                if optim_r == r:
                    break
                else:
                    r = optim_r
        if r >= min_r2:
            return (bp_l, bp_r, r)
        return None

    def _scan(
        self: "PhraseSearcher",
        doc: Doc,
        query: DocLike,
        min_r1: int,
        **kwargs: ty.Any,
    ) -> ty.Optional[ty.Dict[int, int]]:
        """Finds potential match start indices.

        Iterates through `doc` by chunks of `len(query)`
        and comapares each resulting chunk against the `query` looking for
        match ratios >= `min_r1`.

        Args:
            doc: `Doc` object to search over.
            query: `Doc` or `Span` to match against `doc`.
            min_r1: Minimum match ratio required for
                selection during the initial search over `doc`.
                This should be "low" in general because match boundaries
                are not flexed here. A value of `0` here means all chunks of
                `len(query)` in `doc` will have their boundaries flexed and recompared
                during match optimization. Lower `min_r1` will result in more
                fine-grained matching but will run slower.
            **kwargs: Overflow for child keyword arguments.

        Returns:
            Dict of match start index keys to match ratio values or `None`.
        """
        doc_len = len(doc)
        query_len = len(query)
        if not query_len:
            return None
        match_values: ty.Dict[int, int] = dict()
        i = 0
        while i + query_len <= doc_len:
            match = self.compare(
                query,
                doc[i : i + query_len],
                min_r=min_r1 if min_r1 else 1,
                **kwargs,
            )
            if match:
                match_values[i] = match
            i += 1
        if match_values:
            return match_values
        else:
            return None

    @staticmethod
    def _calc_flex(query: DocLike, flex: FlexType) -> int:
        """Returns a valid `flex` value given `query`.

        By default `flex` is set to `len(query) // 2`.

        If `flex` is `'max'`, or an `int` > `len(query)`,
        `flex` will be set to `len(query)` value instead.

        If `flex` is `'min`', or an `int` < `0`,
        `flex` will be set to `0` instead.

        Args:
            query: The `Doc`, `Span`, or `Token` to match with.
            flex: Either `"default"`: `len(query) // 2`,
                `"max"`: `len(query)`,
                `"min"`: `0`,
                or an `int`.

        Returns:
            The new `flex` value.

        Raises:
            TypeError: If `flex` is not `"default"`, `"max"`, `"min"`, or an `int`.

        Warnings:
            FlexWarning:
                If `flex` > `len(query)` or `flex` < `0`.
        """
        if flex == "default":
            flex = len(query) // 2
        elif flex == "max":
            flex = len(query)
        elif flex == "min":
            flex = 0
        elif isinstance(flex, int):
            query_len = len(query)
            if flex > query_len:
                warnings.warn(
                    f"""`flex` of size `{flex}` is > `len(query)`, `{query_len}`.
                        Setting `flex` to `{query_len}` instead.""",
                    FlexWarning,
                    stacklevel=2,
                )
                flex = query_len
            elif flex < 0:
                warnings.warn(
                    """`flex` values < `0` are not allowed.
                    Setting to the min, `0`, instead.""",
                    FlexWarning,
                    stacklevel=2,
                )
                flex = 0
        else:
            raise TypeError(
                (
                    "`flex` must be a `FlexLiteral` (`'default'`,",
                    "`'max'` or `'min'`), or an `int`.",
                )
            )
        return flex

    @staticmethod
    def _set_ratios(
        min_r: int, min_r1: ty.Optional[int], min_r2: ty.Optional[int]
    ) -> ty.Tuple[int, int]:
        """Sets `min_r1` and `min_r2` based on `min_r` heuristic or provided values."""
        min_r1_ = min_r1 if min_r1 is not None else round(min_r / 1.5)
        min_r2_ = min_r2 if min_r2 is not None else min_r
        return min_r1_, min_r2_

    @staticmethod
    def _check_ratios(
        min_r1: int, min_r2: int, thresh: int, flex: int
    ) -> ty.Tuple[int, int, int]:
        """Ensures match ratio requirements are not set to illegal values."""
        if flex:
            if min_r1 > min_r2:
                warnings.warn(
                    "`min_r1` > `min_r2`, setting `min_r1` = `min_r2`",
                    RatioWarning,
                    stacklevel=2,
                )
                min_r1 = min_r2
            if thresh < min_r2:
                warnings.warn(
                    "`thresh` < `min_r2`, setting `thresh` = `min_r2`",
                    RatioWarning,
                    stacklevel=2,
                )
                thresh = min_r2
        else:
            min_r1 = min_r2
        return min_r1, min_r2, thresh



"""Custom spaCy attributes for spaczz."""
import typing as ty
import warnings

from spacy.tokens import Doc
from spacy.tokens import Span
from spacy.tokens import Token

from .customtypes import SpaczzType
from .exceptions import AttrOverwriteWarning


class SpaczzAttrs:
    """Adds spaczz custom attributes to spaCy."""

    _initialized = False

    @classmethod
    def initialize(cls: ty.Type["SpaczzAttrs"]) -> None:
        """Initializes and registers custom attributes."""
        if not cls._initialized:
            try:
                Token.set_extension("spaczz_token", default=False)
                Token.set_extension("spaczz_type", default=None)
                Token.set_extension("spaczz_ratio", default=None)
                Token.set_extension("spaczz_pattern", default=None)

                Span.set_extension("spaczz_ent", getter=cls.get_spaczz_ent)
                Span.set_extension("spaczz_type", getter=cls.get_span_type)
                Span.set_extension("spaczz_types", getter=cls.get_span_types)
                Span.set_extension("spaczz_ratio", getter=cls.get_ratio)
                Span.set_extension("spaczz_pattern", getter=cls.get_pattern)

                Doc.set_extension("spaczz_doc", getter=cls.get_spaczz_doc)
                Doc.set_extension("spaczz_types", getter=cls.get_doc_types)
                cls._initialized = True
            except ValueError:
                warnings.warn(
                    """One or more spaczz custom extensions has already been registered.
                    These are being force overwritten. Please avoid defining personal,
                    custom extensions prepended with `"spaczz_"`.
                """,
                    AttrOverwriteWarning,
                    stacklevel=2,
                )
                Token.set_extension("spaczz_token", default=False, force=True)
                Token.set_extension("spaczz_type", default=None, force=True)
                Token.set_extension("spaczz_ratio", default=None, force=True)

                Span.set_extension("spaczz_type", getter=cls.get_span_type, force=True)
                Span.set_extension(
                    "spaczz_types", getter=cls.get_span_types, force=True
                )
                Span.set_extension("spaczz_ratio", getter=cls.get_ratio, force=True)
                Span.set_extension("spaczz_pattern", getter=cls.get_pattern, force=True)

                Doc.set_extension("spaczz_doc", getter=cls.get_spaczz_doc, force=True)
                Doc.set_extension("spaczz_types", getter=cls.get_doc_types, force=True)

    @staticmethod
    def get_spaczz_ent(span: Span) -> bool:
        """Getter for spaczz_ent `Span` attribute."""
        return all([token._.spaczz_token for token in span])

    @classmethod
    def get_span_type(
        cls: ty.Type["SpaczzAttrs"], span: Span
    ) -> ty.Optional[SpaczzType]:
        """Getter for spaczz_type `Span` attribute."""
        if cls._all_equal([token._.spaczz_type for token in span]):
            return span[0]._.spaczz_type
        else:
            return None

    @staticmethod
    def get_span_types(span: Span) -> ty.Set[SpaczzType]:
        """Getter for spaczz_types `Span` attribute."""
        types = [token._.spaczz_type for token in span if token._.spaczz_type]
        return set(types)

    @classmethod
    def get_ratio(cls: ty.Type["SpaczzAttrs"], span: Span) -> ty.Optional[int]:
        """Getter for spaczz_ratio `Span` attribute."""
        if cls._all_equal([token._.spaczz_ratio for token in span]):
            return span[0]._.spaczz_ratio
        else:
            return None

    @classmethod
    def get_pattern(cls: ty.Type["SpaczzAttrs"], span: Span) -> ty.Optional[str]:
        """Getter for spaczz_pattern `Span` attribute."""
        if cls._all_equal([token._.spaczz_pattern for token in span]):
            return span[0]._.spaczz_pattern
        else:
            return None

    @staticmethod
    def get_spaczz_doc(doc: Doc) -> bool:
        """Getter for spaczz_doc `Doc` attribute."""
        return any([token._.spaczz_token for token in doc])

    @staticmethod
    def get_doc_types(doc: Doc) -> ty.Set[SpaczzType]:
        """Getter for spaczz_types `Doc` attribute."""
        types = [token._.spaczz_type for token in doc if token._.spaczz_type]
        return set(types)

    @staticmethod
    def _all_equal(iterable: ty.Iterable[ty.Any]) -> bool:
        """Tests if all elements of iterable are equal."""
        iterator = iter(iterable)
        try:
            first = next(iterator)
        except StopIteration:
            return True
        return all(first == rest for rest in iterator)



"""Module for TokenMatcher with an API semi-analogous to spaCy's Matcher."""
from copy import deepcopy
import typing as ty

from spacy.matcher import Matcher
from spacy.tokens import Doc
from spacy.vocab import Vocab
import srsly

from .._search import TokenSearcher
from ..customtypes import SpaczzType


class TokenMatcher:
    """spaCy-like matcher for finding fuzzy token matches in `Doc` objects.

    Fuzzy matches added patterns against the `Doc` object it is called on.
    Accepts labeled patterns in the form of lists of dictionaries
    where each list describes an individual pattern and each
    dictionary describes an individual token.

    Uses extended spaCy token matching patterns.
    "FUZZY" and "FREGEX" are the two additional spaCy token pattern options.

    For example::

        [
            {"TEXT": {"FREGEX": "(database){e<=1}"}},
            {"LOWER": {"FUZZY": "access", "MIN_R": 85, "FUZZY_FUNC": "partial"}},
        ]

    Make sure to use uppercase dictionary keys in patterns.

    Attributes:
        name (str): Class attribute - the name of the matcher.
        defaults (dict[str, bool|int|str]):
            Keyword arguments to be used as default match settings.
            Per-pattern match settings take precedence over defaults.

    Match Settings:
        ignore_case (bool): Whether to lower-case text before matching.
            Can only be set at the pattern level. For "FUZZY" and "FREGEX" patterns.
            Default is `True`.
        min_r (int): Minimum match ratio required. For "FUZZY" and "FREGEX" patterns.
        fuzzy_func (str): Key name of fuzzy matching function to use.
            Can only be set at the pattern level. For "FUZZY" patterns only.
            All rapidfuzz matching functions with default settings are available,
            however any token-based functions provide no utility at the individual
            token level. Additional fuzzy matching functions can be registered by users.
            Included, and useful, functions are:

            * `"simple"` = `ratio`
            * `"partial"` = `partial_ratio`
            * `"quick"` = `QRatio`
            * `"partial_alignment"` = `partial_ratio_alignment`
                (Requires `rapidfuzz>=2.0.3`)

            Default is `"simple`".
        fuzzy_weights: Name of weighting method for regex insertion, deletion, and
            substituion counts. Can only be set at the pattern level. For "FREGEX"
            patterns only. Included weighting methods are:

            * `"indel"` = `(1, 1, 2)`
            * `"lev"` = `(1, 1, 1)`

            Default is `"indel"`.
        predef: Whether regex should be interpreted as a key to
            a predefined regex pattern or not. Can only be set at the pattern level.
            For "FREGEX" patterns only. Default is `False`.
    """

    name = "token_matcher"

    def __init__(self: "TokenMatcher", vocab: Vocab, **defaults: ty.Any) -> None:
        """Initializes the matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object. Purely for consistency between spaCy
                and spaczz matcher APIs for now. spaczz matchers are currently pure
                Python and do not share vocabulary with spacy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings for the class instance.
        """
        self.defaults = defaults
        self._type: SpaczzType = "token"
        self._callbacks: ty.Dict[str, TokenCallback] = {}
        self._patterns: ty.DefaultDict[
            str, ty.List[ty.List[ty.Dict[str, ty.Any]]]
        ] = ty.DefaultDict(list)
        self._searcher = TokenSearcher(vocab=vocab)

    def __call__(
        self: "TokenMatcher", doc: Doc
    ) -> ty.List[ty.Tuple[str, int, int, int, str]]:
        """Finds matches in `doc` given the matchers patterns.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of `MatchResult` tuples,
            (label, start index, end index, match ratio, pattern).

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> doc = nlp("Rdley Scot was the director of Alien.")
            >>> matcher.add("NAME", [
                [{"TEXT": {"FUZZY": "Ridley"}},
                {"TEXT": {"FUZZY": "Scott"}}]
                ])
            >>> matcher(doc)[0][:4]
            ('NAME', 0, 2, 90)
        """
        matches: ty.Set[ty.Tuple[str, int, int, int, str]] = set()
        for label, patterns in self._patterns.items():
            for pattern in patterns:
                spaczz_matches = self._searcher.match(doc, pattern, **self.defaults)
                if spaczz_matches:
                    for spaczz_match in spaczz_matches:
                        matcher = Matcher(self.vocab)
                        matcher.add(label, [self._spacyfy(spaczz_match, pattern)])
                        spacy_matches = matcher(doc)
                        for spacy_match in spacy_matches:
                            matches.add(
                                self._calc_ratio(
                                    doc,
                                    pattern=pattern,
                                    spaczz_match=spaczz_match,
                                    spacy_match=ty.cast(
                                        ty.Tuple[int, int, int], spacy_match
                                    ),
                                )
                            )
        sorted_matches = sorted(
            matches, key=lambda x: (-x[1], x[2] - x[1], x[3]), reverse=True
        )
        for i, (label, _start, _end, _ratio, _pattern) in enumerate(sorted_matches):
            on_match = self._callbacks.get(label)
            if on_match:
                on_match(self, doc, i, sorted_matches)
        return sorted_matches

    def __contains__(self: "TokenMatcher", label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return label in self._patterns

    def __len__(self: "TokenMatcher") -> int:
        """The number of labels added to the matcher."""
        return len(self._patterns)

    def __reduce__(
        self: "TokenMatcher",
    ) -> ty.Tuple[ty.Any, ty.Any]:  # Precisely typing this would be really long.
        """Interface for pickling the matcher."""
        data = (
            self.__class__,
            self.vocab,
            self._patterns,
            self._callbacks,
            self.defaults,
        )
        return (unpickle_matcher, data)

    @property
    def labels(self: "TokenMatcher") -> ty.Tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique labels as a tuple of strings.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> matcher.labels
            ('AUTHOR',)
        """
        return tuple(self._patterns.keys())

    @property
    def patterns(self: "TokenMatcher") -> ty.List[ty.Dict[str, ty.Any]]:
        """Get all patterns and match settings that were added to the matcher.

        Returns:
            The patterns and their respective match settings as a list of dicts.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> matcher.patterns == [
                {
                    "label": "AUTHOR",
                    "pattern": [{"TEXT": {"FUZZY": "Kerouac"}}],
                    "type": "token",
                    },
                    ]
            True
        """
        all_patterns = []
        for label, patterns in self._patterns.items():
            for pattern in patterns:
                p = {"label": label, "pattern": pattern, "type": self.type}
                all_patterns.append(p)
        return all_patterns

    @property
    def type(self: "TokenMatcher") -> SpaczzType:
        """Getter for the matchers `SpaczzType`."""
        return self._type

    @property
    def vocab(self: "TokenMatcher") -> Vocab:
        """Getter for the matchers `Vocab`."""
        return self._searcher.vocab

    def add(
        self: "TokenMatcher",
        label: str,
        patterns: ty.List[ty.List[ty.Dict[str, ty.Any]]],
        on_match: "TokenCallback" = None,
    ) -> None:
        """Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of lists of dicts where each list of dicts represent an
        individual pattern and each dictionary represents an individual token.

        Uses extended spaCy token matching patterns.
        "FUZZY" and "FREGEX" are the two additional spaCy token pattern options.

        For example::

            [
                {"TEXT": {"FREGEX": "(database){e<=1}"}},
                {"LOWER": {"FUZZY": "access", "MIN_R": 85, "FUZZY_FUNC": "partial"}},
            ]

        Make sure to use uppercase dictionary keys in patterns.

        Args:
            label: Name of the rule added to the matcher.
            patterns: List of lists of dicts that will be matched
                against the `Doc` object the matcher is called on.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Raises:
            TypeError: If patterns is not a list of `Doc` objects.
            ValueError: Patterns cannot have zero tokens.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> "AUTHOR" in matcher
            True
        """
        for pattern in patterns:
            if len(pattern) == 0:
                raise ValueError("Pattern cannot have zero tokens.")
            if isinstance(pattern, list):
                self._patterns[label].append(list(pattern))
            else:
                raise TypeError("Patterns must be lists of dictionaries.")
        self._callbacks[label] = on_match

    def remove(self: "TokenMatcher", label: str) -> None:
        """Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Raises:
            ValueError: If label does not exist in the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import TokenMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = TokenMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [[{"TEXT": {"FUZZY": "Kerouac"}}]])
            >>> matcher.remove("AUTHOR")
            >>> "AUTHOR" in matcher
            False
        """
        try:
            del self._patterns[label]
            del self._callbacks[label]
        except KeyError:
            raise ValueError(
                f"The label: {label} does not exist within the matcher rules."
            )

    def _calc_ratio(
        self: "TokenMatcher",
        doc: Doc,
        pattern: ty.List[ty.Dict[str, ty.Any]],
        spaczz_match: ty.List[ty.Tuple[str, str, int]],
        spacy_match: ty.Tuple[int, int, int],
    ) -> ty.Tuple[str, int, int, int, str]:
        """Calculates the fuzzy ratio for the entire token match."""
        ratio = round(
            sum(
                [
                    token_match[2]
                    / sum(
                        [len(token) for token in doc[spacy_match[1] : spacy_match[2]]]
                    )
                    * len(token)
                    for token, token_match in zip(  # noqa: B905
                        doc[spacy_match[1] : spacy_match[2]], spaczz_match
                    )
                ]
            )
        )

        return (
            ty.cast(str, self.vocab.strings[spacy_match[0]]),
            spacy_match[1],
            spacy_match[2],
            ratio,
            ty.cast(str, srsly.json_dumps(pattern)),
        )

    @staticmethod
    def _spacyfy(
        match: ty.List[ty.Tuple[str, str, int]],
        pattern: ty.List[ty.Dict[str, ty.Any]],
    ) -> ty.List[ty.Dict[str, ty.Any]]:
        """Turns token searcher matches into spaCy `Matcher` compatible patterns."""
        new_pattern = deepcopy(pattern)
        for i, token in enumerate(match):
            if token[0]:
                del new_pattern[i][token[0]]
                new_pattern[i]["TEXT"] = token[1]
        return new_pattern


TokenCallback = ty.Optional[
    ty.Callable[
        [TokenMatcher, Doc, int, ty.List[ty.Tuple[str, int, int, int, str]]], None
    ]
]


def unpickle_matcher(
    matcher: ty.Type[TokenMatcher],
    vocab: Vocab,
    patterns: ty.DefaultDict[str, ty.List[ty.List[ty.Dict[str, ty.Any]]]],
    callbacks: ty.Dict[str, TokenCallback],
    defaults: ty.Any,
) -> TokenMatcher:
    """Will return a matcher from pickle protocol."""
    matcher_instance = matcher(vocab, **defaults)
    for key, specs in patterns.items():
        callback = callbacks.get(key)
        matcher_instance.add(key, specs, on_match=callback)
    return matcher_instance



"""`RegexSearcher` searches for phrase-based regex matches in spaCy `Doc` objects."""
import typing as ty

try:
    from typing import Match
except ImportError:  # pragma: no cover
    from regex import Match  # type: ignore

from spacy.tokens import Doc
from spacy.tokens import Span
from spacy.vocab import Vocab

from .searchutil import filter_overlapping_matches
from .searchutil import normalize_fuzzy_regex_counts
from .searchutil import parse_regex
from ..customtypes import SearchResult


class RegexSearcher:
    """Class for phrase-based regex match searching in spaCy `Doc` objects."""

    def __init__(
        self: "RegexSearcher",
        vocab: Vocab,
    ) -> None:
        """Initializes the searcher."""
        self.vocab = vocab

    def match(
        self: "RegexSearcher",
        doc: Doc,
        query: str,
        *,
        ignore_case: bool = True,
        min_r: int = 75,
        partial: bool = True,
        predef: bool = False,
        fuzzy_weights: str = "indel",
    ) -> ty.List[SearchResult]:
        """Performs regex matching on a `Doc` object.

        Matches on the character level and then maps matches back
        to tokens. If a character cannot be mapped back to a token it means
        the character is a space tokens are split on, which happens when regex
        matches produce leading or trailing whitespace. Confirm your regex pattern
        will not do this to avoid this issue.

        To utilize regex flags, use inline flags.

        Args:
            doc: `Doc` to search for matches.
            query: A string to be compiled to regex, or the key name of a predefined
                regex pattern.
            ignore_case: Whether to lower-case text before matching.
                Default is `True`.
            min_r: Minimum match ratio required for fuzzy regex matching.
                Default is `75`.
            fuzzy_weights: Name of weighting method for regex insertion, deletion, and
                substituion counts. Default is `"indel"`.
            partial: Whether partial matches should be extended
                to existing span boundaries in doc or not, e.x. the regex only matches
                part of a `Token` or `Span`. Default is `True`.
            predef: Whether regex should be interpreted as a key to
                a predefined regex pattern or not. Default is `False`.

        Returns:
            List of match tuples each containing a start index, end index,
            and match ratio.

        Example:
            >>> import spacy
            >>> from spaczz._search import RegexSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = RegexSearcher(nlp.vocab)
            >>> doc = nlp("My phone number is (555) 555-5555.")
            >>> searcher.match(doc, "phones", predef=True)
            [(4, 10, 100)]
        """
        compiled_regex = parse_regex(query, predef=predef)
        char_to_token_map = self._map_chars_to_tokens(doc)

        regex_matches = [
            self._spans_from_regex(
                doc, match=match, partial=partial, char_to_token_map=char_to_token_map
            )
            for match in compiled_regex.finditer(
                doc.text.lower() if ignore_case else doc.text
            )
        ]

        formatted_matches = [
            (
                regex_match[0].start,
                regex_match[0].end,
                normalize_fuzzy_regex_counts(
                    regex_match[0].text,
                    fuzzy_counts=regex_match[1],
                    fuzzy_weights=fuzzy_weights,
                ),
            )
            for regex_match in regex_matches
            if regex_match
        ]

        return filter_overlapping_matches(
            sorted(
                [
                    formatted_match
                    for formatted_match in formatted_matches
                    if formatted_match[2] >= min_r
                ],
                key=lambda x: (-x[2], x[0]),
            )
        )

    @staticmethod
    def _map_chars_to_tokens(doc: Doc) -> ty.Dict[int, int]:
        """Maps characters to tokens."""
        chars_to_tokens = {}
        for token in doc:
            for i in range(token.idx, token.idx + len(token.text)):
                chars_to_tokens[i] = token.i
        return chars_to_tokens

    @staticmethod
    def _spans_from_regex(
        doc: Doc,
        match: Match[str],
        partial: bool,
        char_to_token_map: ty.Dict[int, int],
    ) -> ty.Optional[ty.Tuple[Span, ty.Tuple[int, int, int]]]:
        """Creates `Span` objects from regex matches."""
        start, end = match.span()
        counts = getattr(match, "fuzzy_counts", (0, 0, 0))
        span = doc.char_span(start, end)
        if span:
            return (span, counts)
        if partial:
            start_token = char_to_token_map.get(start)
            end_token = char_to_token_map.get(end - 1)
            if start_token and end_token:
                span = Span(doc, start_token, end_token + 1)
                return (span, counts)
        return None



"""Module for search utilities."""
import itertools
import typing as ty

import regex as re

from ..customtypes import SearchResult
from ..exceptions import RegexParseError
from ..registry import get_re_pattern
from ..registry import get_re_weights


def filter_overlapping_matches(
    matches: ty.Iterable[SearchResult],
) -> ty.List[SearchResult]:
    """Prevents multiple matches from overlapping.

    Expects matches to be pre-sorted by descending ratio
    then ascending start index.
    If more than one match includes the same tokens
    the first of these matches is kept.

    Args:
        matches: Iterable of matches (start index, end index, ratio tuples).

    Returns:
        The filtered list of matches.

    Example:
        >>> from spaczz._search.searchutil import filter_overlapping_matches
        >>> matches = [(1, 3, 80), (1, 2, 70)]
        >>> filter_overlapping_matches(matches)
        [(1, 3, 80)]
    """
    filtered_matches: ty.List[SearchResult] = []
    for match in matches:
        if not set(range(match[0], match[1])).intersection(
            itertools.chain(*[set(range(n[0], n[1])) for n in filtered_matches])
        ):
            filtered_matches.append(match)
    return filtered_matches


def parse_regex(
    regex_str: str,
    predef: bool = False,
) -> ty.Pattern:
    """Parses a string into a regex pattern.

    Will treat `regex_str` as a key name for a predefined regex if `predef=True`.

    Args:
        regex_str: String to compile into a regex pattern.
        predef: Whether regex should be interpreted as a key to
            a predefined regex pattern or not. Default is `False`.

    Returns:
        A compiled regex pattern.

    Raises:
        RegexParseError: If regex compilation produces any errors.

    Example:
        >>> from spaczz._search.searchutil import parse_regex
        >>> pattern = parse_regex("Test")
        >>> isinstance(pattern, re.Pattern)
        True
    """
    if predef:
        return get_re_pattern(regex_str)
    try:
        return re.compile(
            regex_str,
        )
    except (re._regex_core.error, TypeError, ValueError) as e:
        raise RegexParseError(e)


def normalize_fuzzy_regex_counts(
    match: str, fuzzy_counts: ty.Tuple[int, int, int], fuzzy_weights: str
) -> int:
    """Normalizes fuzzy regex counts to a fuzzy ratio."""
    if fuzzy_counts == (0, 0, 0):
        return 100

    weights = get_re_weights(fuzzy_weights)
    s1_len = len(match) - fuzzy_counts[1] + fuzzy_counts[2]
    s2_len = len(match)
    fuzzy_counts_ = (
        a * b
        for a, b in zip(  # noqa: B905
            (fuzzy_counts[1], fuzzy_counts[2], fuzzy_counts[0]), weights
        )
    )

    if weights[2] <= weights[0] + weights[1]:
        dist_max = min(s1_len, s2_len) * weights[2]
    else:
        dist_max = s1_len * weights[1] + s2_len * weights[0]

    if s1_len > s2_len:
        dist_max += (s1_len - s2_len) * weights[1]
    elif s1_len < s2_len:
        dist_max += (s2_len - s1_len) * weights[0]

    r = 100 * sum(fuzzy_counts_) / dist_max
    return round(100 - r)



"""Registry of commonly used fuzzy weights based on IDS counts."""
from functools import lru_cache

import catalogue

re_weights = catalogue.create("spaczz", "re_weights", entry_points=True)

re_weights.register("indel", func=(1, 1, 2))
re_weights.register("lev", func=(1, 1, 1))

get_re_weights = lru_cache(None)(re_weights.get)



"""Module for TokenSearcher: flexible token searching in spaCy `Doc` objects."""
import itertools
import typing as ty

import regex
from spacy.tokens import Doc
from spacy.tokens import Token
from spacy.vocab import Vocab

from .searchutil import normalize_fuzzy_regex_counts
from .searchutil import parse_regex
from ..registry import get_fuzzy_func


class TokenSearcher:
    """Class for token-based fuzzy match searching in spaCy `Doc` objects."""

    def __init__(self: "TokenSearcher", vocab: Vocab) -> None:
        """Initializes the searcher."""
        self.vocab = vocab

    def match(
        self: "TokenSearcher",
        doc: Doc,
        pattern: ty.List[ty.Dict[str, ty.Any]],
        min_r: int = 75,
    ) -> ty.List[ty.List[ty.Tuple[str, str, int]]]:
        """Finds potential token pattern matches in a `Doc` object.

        Make sure to use uppercase dictionary keys in patterns.

        Args:
            doc: `Doc` object to search for matches.
            pattern: Individual spaCy token pattern.
            min_r: Minimum match ratio required for fuzzy matching.
                Can be overwritten with pattern-level match settings.
                Default is `75`.

        Returns:
            A list of lists with each inner list representing a potential match.
            The inner lists will be populated with key, value, match ratio tuples for
            matched fuzzy token patterns. Non-fuzzy token patterns will have an empty
            string key and value, and a match ratio of `100`.

        Example:
            >>> import spacy
            >>> from spaczz._search import TokenSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = TokenSearcher(nlp)
            >>> doc = nlp("I was prescribed zithramax and advar")
            >>> pattern = [
                {"TEXT": {"FUZZY": "zithromax"}},
                {"POS": "CCONJ"},
                {"TEXT": {"FREGEX": "(advair){e<=1}"}}
                ]
            >>> searcher.match(doc, pattern)
            [[('TEXT', 'zithramax', 89), ('', '', 100), ('TEXT', 'advar', 91)]]
        """
        matches = []
        matches = [
            self._iter_pattern(seq, pattern, min_r=min_r)
            for seq in self._n_wise(doc, len(pattern))
        ]
        return [
            match
            for i, match in enumerate(matches)
            if match and match not in matches[:i]
        ]

    @staticmethod
    def fuzzy_compare(
        s1: str,
        s2: str,
        *,
        ignore_case: bool = True,
        min_r: int = 0,
        fuzzy_func: str = "simple",
    ) -> int:
        """Peforms fuzzy matching between two strings.

        Applies the given fuzzy matching algorithm (`fuzzy_func`)
        to two strings and returns the resulting fuzzy ratio.

        Args:
            s1: First string for comparison.
            s2: Second string for comparison.
            ignore_case: Whether to lower-case `s1` and `s2`
                before comparison or not. Default is `True`.
            min_r: Minimum ratio needed to match as a value between `0` and `100`.
                For ratio < min_r, `0` is returned instead.
                Default is `0`, which deactivates this behaviour.
            fuzzy_func: Key name of fuzzy matching function to use.

        Returns:
            The fuzzy ratio between `s1` and `s2`.

        Example:
            >>> import spacy
            >>> from spaczz._search import TokenSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = TokenSearcher(nlp.vocab)
            >>> searcher.fuzzy_compare("spaczz", "spacy")
            73
        """
        if ignore_case:
            s1 = s1.lower()
            s2 = s2.lower()

        return round(get_fuzzy_func(fuzzy_func)(s1, s2, score_cutoff=min_r))

    @staticmethod
    def regex_compare(
        text: str,
        pattern: str,
        *,
        ignore_case: bool = False,
        min_r: int = 0,
        fuzzy_weights: str = "indel",
        predef: bool = False,
    ) -> int:
        """Performs fuzzy-regex supporting regex matching between two strings.

        Args:
            text: The string to match against.
            pattern: The regex pattern string.
            ignore_case: Whether to lower-case text before matching.
                Default is `True`.
            min_r: Minimum match ratio required for fuzzy regex matching.
                Default is `75`.
            fuzzy_weights: Name of weighting method for regex insertion, deletion, and
                substituion counts.
            predef: Whether regex should be interpreted as a key to
                a predefined regex pattern or not. Default is `False`.

        Returns:
            The fuzzy ratio of the pattern to the text based on insertions,
            substitutions, and deletions.

        Example:
            >>> import spacy
            >>> from spaczz._search import TokenSearcher
            >>> nlp = spacy.blank("en")
            >>> searcher = TokenSearcher(nlp)
            >>> searcher.regex_compare("sequel", "(sql){i<=3}")
            67
        """
        pattern_ = parse_regex(pattern, predef=predef)
        if ignore_case:
            text = text.lower()
        match = regex.match(pattern_, text)
        if match:
            r = normalize_fuzzy_regex_counts(
                match.group(0),
                fuzzy_counts=getattr(match, "fuzzy_counts", (0, 0, 0)),
                fuzzy_weights=fuzzy_weights,
            )
            if r >= min_r:
                return r
        return 0

    def _iter_pattern(
        self: "TokenSearcher",
        seq: ty.Tuple[Token, ...],
        pattern: ty.List[ty.Dict[str, ty.Any]],
        min_r: int,
    ) -> ty.List[ty.Tuple[str, str, int]]:
        """Evaluates each token in a pattern against a doc token sequence."""
        seq_matches: ty.List[ty.Tuple[str, str, int]] = []
        for i, token in enumerate(pattern):
            pattern_dict, case, case_bool = self._parse_case(token)
            if isinstance(pattern_dict, dict):
                pattern_text, pattern_type = self._parse_type(pattern_dict)
                if pattern_text and pattern_type == "FUZZY":
                    r = self.fuzzy_compare(
                        seq[i].text,
                        pattern_text,
                        ignore_case=case_bool,
                        min_r=pattern_dict.get("MIN_R", min_r),
                        fuzzy_func=pattern_dict.get("FUZZY_FUNC", "simple"),
                    )
                    if r:
                        seq_matches.append((case, seq[i].text, r))
                    else:
                        return []
                elif pattern_text and pattern_type == "FREGEX":
                    r = self.regex_compare(
                        seq[i].text,
                        pattern_text,
                        predef=pattern_dict.get("PREDEF", False),
                        ignore_case=case_bool,
                        min_r=pattern_dict.get("MIN_R", min_r),
                        fuzzy_weights=pattern_dict.get("FUZZY_WEIGHTS", "indel"),
                    )
                    if r:
                        seq_matches.append((case, seq[i].text, r))
                    else:
                        return []
                else:
                    seq_matches.append(("", "", 100))
            else:
                seq_matches.append(("", "", 100))
        return seq_matches

    @staticmethod
    def _parse_case(
        token: ty.Dict[str, ty.Any]
    ) -> ty.Tuple[ty.Union[str, ty.Dict[ty.Any, ty.Any], None], str, bool]:
        """Parses the case of a token pattern."""
        text = token.get("TEXT")
        if text:
            return text, "TEXT", False
        return token.get("LOWER"), "LOWER", True

    @staticmethod
    def _parse_type(pattern_dict: ty.Dict[str, ty.Any]) -> ty.Tuple[str, str]:
        """Parses the type of a token pattern."""
        fuzzy_text = pattern_dict.get("FUZZY")
        regex_text = pattern_dict.get("FREGEX")
        if fuzzy_text:
            return fuzzy_text, "FUZZY"
        elif regex_text:
            return regex_text, "FREGEX"
        return "", ""

    @staticmethod
    def _n_wise(iterable: ty.Iterable[ty.Any], n: int) -> ty.Iterator[ty.Any]:
        """Iterates over an iterable in slices of length n by one step at a time."""
        iterables = itertools.tee(iterable, n)
        for i in range(len(iterables)):
            for _ in range(i):
                next(iterables[i], None)
        return zip(*iterables)  # noqa B905



"""Module for RegexMatcher with an API semi-analogous to spaCy's `PhraseMatcher`."""
import typing as ty
import warnings

from spacy.tokens import Doc
from spacy.vocab import Vocab

from .._search import RegexSearcher
from ..customtypes import MatchResult
from ..customtypes import SpaczzType
from ..exceptions import KwargsWarning
from ..util import nest_defaultdict


class RegexMatcher:
    """spaCy-like matcher for finding regex phrase matches in `Doc` objects.

    Regex matches patterns against the `Doc` it is called on.
    Accepts labeled patterns in the form of strings with optional,
    per-pattern match settings.

    To utilize regex flags, use inline flags.

    Attributes:
        name (str): Class attribute - the name of the matcher.
        defaults (dict[str, bool|int|str]):
            Keyword arguments to be used as default match settings.
            Per-pattern match settings take precedence over defaults.

    Match Settings:
        ignore_case (bool): Whether to lower-case text before matching.
            Default is `True`.
        min_r (int): Minimum match ratio required.
        fuzzy_weights (str): Name of weighting method for regex insertion, deletion, and
            substituion counts. Additional weighting methods can be registered
            by users. Included weighting methods are:

            * `"indel"` = `(1, 1, 2)`
            * `"lev"` = `(1, 1, 1)`

            Default is `"indel"`.
        partial: (bool): Whether partial matches should be extended
                to `Token` or `Span` boundaries in `doc` or not.
                For example, the regex only matches part of a `Token` or `Span` in
                `doc`. Default is `True`.
        predef (string): Whether the regex string should be interpreted as a key to
                a predefined regex pattern or not. Additional predefined regex patterns
                can be registered by users. The included predefined regex patterns are:

                * `"dates"`
                * `"times"`
                * `"phones"`
                * `"phones_with_exts"`
                * `"links"`
                * `"emails"`
                * `"ips"`
                * `"ipv6s"`
                * `"prices"`
                * `"hex_colors"`
                * `"credit_cards"`
                * `"btc_addresses"`
                * `"street_addresses"`
                * `"zip_codes"`
                * `"po_boxes"`
                * `"ssn_numbers"`

                Default is `False`.
    """

    name = "regex_matcher"

    def __init__(
        self: "RegexMatcher",
        vocab: Vocab,
        **defaults: ty.Any,
    ) -> None:
        """Initializes the matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object. Purely for consistency between spaCy
                and spaczz matcher APIs for now. spaczz matchers are currently pure
                Python and do not share vocabulary with spacy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings for the class instance.
        """
        self.defaults = defaults
        self._type: SpaczzType = "regex"
        self._callbacks: ty.Dict[str, RegexCallback] = {}
        self._patterns: ty.DefaultDict[
            str, ty.DefaultDict[str, ty.Any]
        ] = nest_defaultdict(list)
        self._searcher = RegexSearcher(vocab=vocab)

    def __call__(self: "RegexMatcher", doc: Doc) -> ty.List[MatchResult]:
        r"""Finds matches in `doc` given the matchers patterns.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of `MatchResult` tuples,
            (label, start index, end index, match ratio, pattern).

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> doc = nlp.make_doc("I live in the united states, or the US")
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> matcher(doc)[0]
            ('GPE', 4, 6, 100, '[Uu](nited|\\.?) ?[Ss](tates|\\.?)')
        """
        matches = set()
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(  # noqa B905
                patterns["patterns"], patterns["kwargs"]
            ):
                if not kwargs:
                    kwargs = self.defaults
                matches_wo_label = self._searcher.match(doc, pattern, **kwargs)
                if matches_wo_label:
                    matches_w_label = [
                        (label, *match_wo_label, str(pattern))
                        for match_wo_label in matches_wo_label
                    ]
                    for match in matches_w_label:
                        matches.add(match)
        sorted_matches = sorted(
            matches, key=lambda x: (-x[1], x[2] - x[1], x[3]), reverse=True
        )
        for i, (label, _start, _end, _ratio, _pattern) in enumerate(sorted_matches):
            on_match = self._callbacks.get(label)
            if on_match:
                on_match(self, doc, i, sorted_matches)
        return sorted_matches

    def __contains__(self: "RegexMatcher", label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return label in self._patterns

    def __len__(self: "RegexMatcher") -> int:
        """The number of labels added to the matcher."""
        return len(self._patterns)

    def __reduce__(
        self: "RegexMatcher",
    ) -> ty.Tuple[ty.Any, ty.Any]:  # Precisely typing this would be really long.
        """Interface for pickling the matcher."""
        data = (
            self.__class__,
            self.vocab,
            self._patterns,
            self._callbacks,
            self.defaults,
        )
        return (unpickle_matcher, data)

    @property
    def labels(self: "RegexMatcher") -> ty.Tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique labels as a tuple of strings.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("ZIP", ["zip_codes"], [{"predef": True}])
            >>> matcher.labels
            ('ZIP',)
        """
        return tuple(self._patterns.keys())

    @property
    def patterns(self: "RegexMatcher") -> ty.List[ty.Dict[str, ty.Any]]:
        """Get all patterns and match settings that were added to the matcher.

        Returns:
            The patterns and their respective match settings as a list of dicts.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("ZIP", ["zip_codes"], [{"predef": True}])
            >>> matcher.patterns == [
                {
                    "label": "ZIP",
                    "pattern": "zip_codes",
                    "type": "regex",
                    "kwargs": {"predef": True},
                    }
                    ]
            True
        """
        all_patterns = []
        for label, patterns in self._patterns.items():
            for pattern, kwargs in zip(  # noqa: B905
                patterns["patterns"], patterns["kwargs"]
            ):
                p = {"label": label, "pattern": pattern, "type": "regex"}
                if kwargs:
                    p["kwargs"] = kwargs
                all_patterns.append(p)
        return all_patterns

    @property
    def type(self: "RegexMatcher") -> SpaczzType:
        """Getter for the matchers `SpaczzType`."""
        return self._type

    @property
    def vocab(self: "RegexMatcher") -> Vocab:
        """Getter for the matchers `Vocab`."""
        return self._searcher.vocab

    def add(
        self: "RegexMatcher",
        label: str,
        patterns: ty.List[str],
        kwargs: ty.Optional[ty.List[ty.Dict[str, ty.Any]]] = None,
        on_match: "RegexCallback" = None,
    ) -> None:
        r"""Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of `Doc` objects and if `kwargs` is not `None`,
        `kwargs` must be a list of dicts.

        Args:
            label: Name of the rule added to the matcher.
            patterns: `Doc` objects that will be matched
                against the `Doc` object the matcher is called on.
            kwargs: Optional settings to modify the matching behavior.
                If supplying `kwargs`, one per pattern should be included.
                Empty dicts will use the matcher instances default settings.
                Default is `None`.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Raises:
            TypeError: If `patterns` is not a list of strings.
            TypeError: If `kwargs` is not a list of dictionaries.

        Warnings:
            KwargsWarning:
                * If there are more patterns than kwargs
                  default matching settings will be used
                  for extra patterns.
                * If there are more kwargs dicts than patterns,
                  the extra kwargs will be ignored.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> "GPE" in matcher
            True
        """
        if kwargs is None:
            kwargs = [{} for _ in patterns]
        elif len(kwargs) < len(patterns):
            warnings.warn(
                """There are more patterns then there are kwargs.\n
                    Patterns not matched to a kwarg dict will have default settings.""",
                KwargsWarning,
                stacklevel=2,
            )
            kwargs.extend([{} for _ in range(len(patterns) - len(kwargs))])
        elif len(kwargs) > len(patterns):
            warnings.warn(
                """There are more kwargs dicts than patterns.\n
                    The extra kwargs will be ignored.""",
                KwargsWarning,
                stacklevel=2,
            )
        if not isinstance(patterns, list):
            raise TypeError("Patterns must be a list strings.")
        for pattern, kwarg in zip(patterns, kwargs):  # noqa: B905
            if isinstance(pattern, str):
                self._patterns[label]["patterns"].append(pattern)
            else:
                raise TypeError("Patterns must be a list of strings.")
            if isinstance(kwarg, dict):
                self._patterns[label]["kwargs"].append(kwarg)
            else:
                raise TypeError("Kwargs must be a list of dicts.")
        self._callbacks[label] = on_match

    def remove(self: "RegexMatcher", label: str) -> None:
        r"""Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Raises:
            ValueError: If `label` does not exist in the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import RegexMatcher
            >>> nlp = spacy.blank("en")
            >>> matcher = RegexMatcher(nlp.vocab)
            >>> matcher.add("GPE", ["[Uu](nited|\.?) ?[Ss](tates|\.?)"])
            >>> matcher.remove("GPE")
            >>> "GPE" in matcher
            False
        """
        try:
            del self._patterns[label]
            del self._callbacks[label]
        except KeyError:
            raise ValueError(
                f"The label: {label} does not exist within the matcher rules."
            )


RegexCallback = ty.Optional[
    ty.Callable[[RegexMatcher, Doc, int, ty.List[MatchResult]], None]
]


def unpickle_matcher(
    matcher: ty.Type[RegexMatcher],
    vocab: Vocab,
    patterns: ty.DefaultDict[str, ty.DefaultDict[str, ty.Any]],
    callbacks: ty.Dict[str, RegexCallback],
    defaults: ty.Any,
) -> RegexMatcher:
    """Will return a matcher from pickle protocol."""
    matcher_instance = matcher(vocab, **defaults)
    for key, specs in patterns.items():
        callback = callbacks.get(key)
        matcher_instance.add(key, specs["patterns"], specs["kwargs"], on_match=callback)
    return matcher_instance



"""spaczz library for fuzzy matching and extended regex functionality with spaCy."""
try:
    from importlib.metadata import PackageNotFoundError  # type: ignore
    from importlib.metadata import version  # type: ignore
except ImportError:  # pragma: no cover
    from importlib_metadata import version, PackageNotFoundError  # type: ignore

from spaczz.customattrs import SpaczzAttrs

SpaczzAttrs.initialize()

try:
    __version__ = version(__name__)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"



"""`SimilarityMatcher` with an API semi-analogous to spaCy's `PhraseMatcher`."""
import typing as ty

from spacy.tokens import Doc
from spacy.vocab import Vocab

from ._phrasematcher import PhraseCallback
from ._phrasematcher import PhraseMatcher
from .._search import SimilaritySearcher
from ..customtypes import MatchResult
from ..customtypes import SpaczzType


class SimilarityMatcher(PhraseMatcher):
    """spaCy-like matcher for finding phrase similarity matches in `Doc` objects.

    Similarity matches patterns against the `Doc` it is called on.
    Accepts labeled patterns in the form of `Doc` objects with optional,
    per-pattern match settings.

    Similarity matching uses spaCy word vectors if available,
    therefore spaCy vocabs without word vectors may not produce
    useful results. The spaCy medium and large English models provide
    word vectors that will work for this purpose.

    Searching over/with `Doc` objects that do not have vectors will always return
    a similarity score of 0.

    Warnings from spaCy about the above two scenarios are suppressed
    for convenience. However, spaczz will still warn about the former.

    Attributes:
        name (str): Class attribute - the name of the matcher.
        defaults (dict[str, bool|int|str|Literal['default', 'min', 'max']]):
            Keyword arguments to be used as default match settings.
            Per-pattern match settings take precedence over defaults.

    Match Settings:
        ignore_case (bool): Whether to lower-case text before fuzzy matching.
            Default is `True`.
        min_r (int): Minimum match ratio required.
        thresh (int): If this ratio is exceeded in initial scan,
            and `flex > 0`, no optimization will be attempted.
            If `flex == 0`, `thresh` has no effect. Default is `100`.
        flex (int|Literal['default', 'min', 'max']): Number of tokens to move match
            boundaries left and right during optimization. Can be an `int` with a max of
            `len(pattern)` and a min of `0`, (will warn and change if higher or lower).
            `"max"`, `"min"`, or `"default"` are also valid.
            Default is `"default"`: `len(pattern) // 2`.
        min_r1 (int|None): Optional granular control over the minimum match ratio
            required for selection during the initial scan.
            If `flex == 0`, `min_r1` will be overwritten by `min_r2`.
            If `flex > 0`, `min_r1` must be lower than `min_r2` and "low" in general
            because match boundaries are not flexed initially.
            Default is `None`, which will result in `min_r1` being set to
            `round(min_r / 1.5)`.
        min_r2 (int|None): Optional granular control over the minimum match ratio
            required for selection during match optimization.
            Needs to be higher than `min_r1` and "high" in general to ensure only
            quality matches are returned. Default is `None`, which will result in
            `min_r2` being set to `min_r`.

    Warnings:
        MissingVectorsWarning:
            If `vocab` does not contain any word vectors.
    """

    name = "similarity_matcher"

    def __init__(self: "SimilarityMatcher", vocab: Vocab, **defaults: ty.Any) -> None:
        """Initializes the matcher with the given defaults.

        Args:
            vocab: A spacy `Vocab` object. Purely for consistency between spaCy
                and spaczz matcher APIs for now. spaczz matchers are currently pure
                Python and do not share vocabulary with spacy pipelines.
            **defaults: Keyword arguments that will
                be used as default matching settings for the class instance.
        """
        super().__init__(vocab=vocab, **defaults)
        self._type: SpaczzType = "similarity"
        self._searcher = self._get_searcher(vocab)

    @staticmethod
    def _get_searcher(vocab: Vocab) -> SimilaritySearcher:
        """Initializes the searcher for this matcher."""
        return SimilaritySearcher(vocab)

    def __call__(self: "SimilarityMatcher", doc: Doc) -> ty.List[MatchResult]:
        """Finds matches in `doc` given the matchers patterns.

        Args:
            doc: The `Doc` object to match over.

        Returns:
            A list of `MatchResult` tuples,
            (label, start index, end index, match ratio, pattern).

        Example:
            >>> import spacy
            >>> from spaczz.matcher import SimilarityMatcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> matcher = SimilarityMatcher(nlp.vocab)
            >>> doc = nlp("I like apples.")
            >>> matcher.add("FRUIT", [nlp("fruit")], [{'min_r': 60}])
            >>> matcher(doc)
            [('FRUIT', 2, 3, 70, 'fruit')]
        """
        return super().__call__(doc)

    def __contains__(self: "SimilarityMatcher", label: str) -> bool:
        """Whether the matcher contains patterns for a label."""
        return super().__contains__(label)

    def __len__(self: "SimilarityMatcher") -> int:
        """The number of labels added to the matcher."""
        return super().__len__()

    def __reduce__(
        self: "SimilarityMatcher",
    ) -> ty.Tuple[ty.Any, ty.Any]:  # Precisely typing this would be really long.
        """Interface for pickling the matcher."""
        return super().__reduce__()

    @property
    def labels(self: "SimilarityMatcher") -> ty.Tuple[str, ...]:
        """All labels present in the matcher.

        Returns:
            The unique labels as a tuple of strings.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import SimilarityMatcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> matcher = SimilarityMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [nlp("Kerouac")])
            >>> matcher.labels
            ('AUTHOR',)
        """
        return super().labels

    @property
    def patterns(self: "SimilarityMatcher") -> ty.List[ty.Dict[str, ty.Any]]:
        """Get all patterns and match settings that were added to the matcher.

        Returns:
            The patterns and their respective match settings as a list of dicts.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import SimilarityMatcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> matcher = SimilarityMatcher(nlp.vocab)
            >>> matcher.add("AUTHOR", [nlp("Kerouac")],
                [{"ignore_case": False}])
            >>> matcher.patterns == [
                {
                    "label": "AUTHOR",
                    "pattern": "Kerouac",
                    "type": "similarity",
                    "kwargs": {"ignore_case": False}
                    },
                    ]
            True
        """
        return super().patterns

    @property
    def type(self: "SimilarityMatcher") -> SpaczzType:
        """Getter for the matchers `SpaczzType`."""
        return super().type

    @property
    def vocab(self: "SimilarityMatcher") -> Vocab:
        """Getter for the matchers `Vocab`."""
        return super().vocab

    def add(
        self: "SimilarityMatcher",
        label: str,
        patterns: ty.List[Doc],
        kwargs: ty.Optional[ty.List[ty.Dict[str, ty.Any]]] = None,
        on_match: PhraseCallback = None,
    ) -> None:
        """Add a rule to the matcher, consisting of a label and one or more patterns.

        Patterns must be a list of `Doc` objects and if `kwargs` is not `None`,
        `kwargs` must be a list of dicts.

        Args:
            label: Name of the rule added to the matcher.
            patterns: `Doc` objects that will be matched
                against the `Doc` object the matcher is called on.
            kwargs: Optional settings to modify the matching behavior.
                If supplying `kwargs`, one per pattern should be included.
                Empty dicts will use the matcher instances default settings.
                Default is `None`.
            on_match: Optional callback function to modify the
                `Doc` object the matcher is called on after matching.
                Default is `None`.

        Warnings:
            KwargsWarning:
                * If there are more patterns than kwargs
                  default matching settings will be used
                  for extra patterns.
                * If there are more kwargs dicts than patterns,
                  the extra kwargs will be ignored.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import SimilarityMatcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> matcher = SimilarityMatcher(nlp.vocab)
            >>> matcher.add("SOUND", [nlp("mooo")])
            >>> "SOUND" in matcher
            True
        """
        super().add(label, patterns=patterns, kwargs=kwargs, on_match=on_match)

    def remove(self: "SimilarityMatcher", label: str) -> None:
        """Remove a label and its respective patterns from the matcher.

        Args:
            label: Name of the rule added to the matcher.

        Example:
            >>> import spacy
            >>> from spaczz.matcher import SimilarityMatcher
            >>> nlp = spacy.load("en_core_web_md")
            >>> matcher = SimilarityMatcher(nlp.vocab)
            >>> matcher.add("SOUND", [nlp("mooo")])
            >>> matcher.remove("SOUND")
            >>> "SOUND" in matcher
            False
        """
        super().remove(label)



"""Module for various utility functions."""
from collections import defaultdict
from functools import partial
import itertools
from os import PathLike
from pathlib import Path
import typing as ty


def nest_defaultdict(
    default_factory: ty.Any, depth: int = 1, *args: ty.Any, **kwargs: ty.Any
) -> ty.DefaultDict[ty.Any, ty.Any]:
    """Nests defaultdicts where depth nesting is `defaultdict[default_factory]`."""
    result = partial(defaultdict, default_factory)
    for _ in itertools.repeat(None, depth):
        result = partial(defaultdict, result)
    return result(*args, **kwargs)


def ensure_path(path: ty.Union[str, PathLike]) -> Path:
    """Ensure str or Pathlike is converted to a Path.

    Args:
        path: str or PathLike, if it's not a Path, it's converted to Path.

    Returns:
        Path.
    """
    if not isinstance(path, Path):
        return Path(path)
    else:
        return path


def read_from_disk(
    path: ty.Union[str, PathLike], readers: ty.Any, exclude: ty.Any
) -> Path:
    """Reads a pipeline component from disk."""
    path = ensure_path(path)
    for key, reader in readers.items():
        # Split to support file names like meta.json
        if key.split(".")[0] not in exclude:
            reader(path / key)
    return path


def write_to_disk(
    path: ty.Union[str, PathLike], writers: ty.Any, exclude: ty.Any
) -> Path:
    """Writes a pipeline component to disk."""
    path = ensure_path(path)
    if not path.exists():
        path.mkdir()
    for key, writer in writers.items():
        # Split to support file names like meta.json
        if key.split(".")[0] not in exclude:
            writer(path / key)
    return path
